from typing import Any, Dict

from numpy import diag_indices_from, dot, exp, eye, inf, log, tril_indices_from, zeros
from optimix import Function, Vector

from .._util import format_function


class FreeFormCov(Function):
    """
    General definite positive matrix, K = LLᵀ + ϵI.

    A d×d covariance matrix K will have ((d+1)⋅d)/2 parameters defining the lower
    triangular elements of a Cholesky matrix L such that:

        K = LLᵀ + ϵI,

    for a very small positive number ϵ. That additional term is necessary to avoid
    singular and ill conditioned covariance matrices.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import FreeFormCov
        >>>
        >>> cov = FreeFormCov(2)
        >>> cov.L = [[1., .0], [0.5, 2.]]
        >>> print(cov.gradient()["Lu"])
        [[[0.  2.  0. ]
          [1.  0.5 0. ]]
        <BLANKLINE>
         [[1.  0.5 0. ]
          [1.  0.  8. ]]]
        >>> cov.name = "K"
        >>> print(cov)
        FreeFormCov(dim=2): K
          L: [[1.  0. ]
              [0.5 2. ]]
    """

    def __init__(self, dim):
        """
        Constructor.

        Parameters
        ----------
        dim : int
            Dimension d of the free-form covariance matrix.
        """
        from numpy_sugar import epsilon

        dim = int(dim)
        tsize = ((dim + 1) * dim) // 2
        self._L = zeros((dim, dim))
        self._tril1 = tril_indices_from(self._L, k=-1)
        self._diag = diag_indices_from(self._L)
        self._L[self._tril1] = 1
        self._L[self._diag] = 0
        self._epsilon = epsilon.small * 1000
        self._Lu = Vector(zeros(tsize))
        self._Lu.value[: tsize - dim] = 1
        n = self.L.shape[0]
        self._grad_Lu = zeros((n, n, self._Lu.shape[0]))
        Function.__init__(self, "FreeCov", Lu=self._Lu)
        bounds = [(-inf, +inf)] * (tsize - dim)
        bounds += [(log(epsilon.small * 1000), +11)] * dim
        self._Lu.bounds = bounds
        self._cache: Dict[str, Any] = {"eig": None}
        self.listen(self._parameters_update)
        self._nparams = tsize

    def _parameters_update(self):
        self._cache["eig"] = None

    @property
    def nparams(self):
        """
        Number of parameters.
        """
        return self._nparams

    def listen(self, func):
        """
        Listen to parameters change.

        Parameters
        ----------
        func : callable
            Function to be called when a parameter changes.
        """
        self._Lu.listen(func)

    @property
    def shape(self):
        """
        Array shape.
        """
        n = self._L.shape[0]
        return (n, n)

    def fix(self):
        """
        Disable parameter optimisation.
        """
        self._Lu.fix()

    def unfix(self):
        """
        Enable parameter optimisation.
        """
        self._Lu.unfix()

    def eigh(self):
        """
        Eigen decomposition of K.

        Returns
        -------
        S : ndarray
            The eigenvalues in ascending order, each repeated according to its
            multiplicity.
        U : ndarray
            Normalized eigenvectors.
        """
        from numpy.linalg import svd

        if self._cache["eig"] is not None:
            return self._cache["eig"]

        U, S = svd(self.L)[:2]
        S *= S
        S += self._epsilon
        self._cache["eig"] = S, U

        return self._cache["eig"]

    @property
    def Lu(self):
        """
        Lower-triangular, flat part of L.
        """
        return self._Lu.value

    @Lu.setter
    def Lu(self, v):
        self._Lu.value = v

    @property
    def L(self):
        """
        Lower-triangular matrix L such that K = LLᵀ + ϵI.

        Returns
        -------
        L : (d, d) ndarray
            Lower-triangular matrix.
        """
        m = len(self._tril1[0])
        self._L[self._tril1] = self._Lu.value[:m]
        self._L[self._diag] = exp(self._Lu.value[m:])
        return self._L

    @L.setter
    def L(self, value):
        self._L[:] = value
        m = len(self._tril1[0])
        self._Lu.value[:m] = self._L[self._tril1]
        self._Lu.value[m:] = log(self._L[self._diag])

    def logdet(self):
        """
        Log of ｜K｜.

        Returns
        -------
        float
            Log-determinant of K.
        """
        from numpy.linalg import slogdet

        K = self.value()

        sign, logdet = slogdet(K)
        if sign != 1.0:
            msg = "The estimated determinant of K is not positive: "
            msg += f" ({sign}, {logdet})."
            raise RuntimeError(msg)

        return logdet

    def value(self):
        """
        Covariance matrix.

        Returns
        -------
        K : ndarray
            Matrix K = LLᵀ + ϵI, for a very small positive number ϵ.
        """
        K = dot(self.L, self.L.T)
        return K + self._epsilon * eye(K.shape[0])

    def gradient(self):
        """
        Derivative of the covariance matrix over the parameters of L.

        Returns
        -------
        Lu : ndarray
            Derivative of K over the lower triangular part of L.
        """
        L = self.L
        self._grad_Lu[:] = 0

        for i in range(len(self._tril1[0])):
            row = self._tril1[0][i]
            col = self._tril1[1][i]
            self._grad_Lu[row, :, i] = L[:, col]
            self._grad_Lu[:, row, i] += L[:, col]

        m = len(self._tril1[0])
        for i in range(len(self._diag[0])):
            row = self._diag[0][i]
            col = self._diag[1][i]
            self._grad_Lu[row, :, m + i] = L[row, col] * L[:, col]
            self._grad_Lu[:, row, m + i] += L[row, col] * L[:, col]

        return {"Lu": self._grad_Lu}

    def __str__(self):
        return format_function(self, {"dim": self._L.shape[0]}, [("L", self.L)])



from numpy import asarray, block, clip, inf, kron, sqrt
from numpy.linalg import pinv

from glimix_core._util import rsolve, unvec, vec

from .._util import cached_property, log2pi, lu_solve, safe_log


class KronFastScanner:
    """
    Approximated fast inference over several covariates.

    Specifically, it maximizes the log of the marginal likelihood ::

        log(p(Y)ⱼ) = log𝓝(vec(Y) | (A ⊗ X)vec(𝚩ⱼ) + (Aⱼ ⊗ Xⱼ)vec(𝚨ⱼ), sⱼK),

    where K = C₀ ⊗ GGᵀ + C₁ ⊗ I and ⱼ index the candidates set. For performance purpose,
    we optimise only the fixed-effect sizes and scale parameters. Therefore, K is fixed
    throughout the process.
    """

    def __init__(self, Y, A, X, G, terms):
        """
        Constructor.

        Parameters
        ----------
        Y : (n, p) array_like
            Outcome matrix.
        A : (n, n) array_like
            Trait-by-trait design matrix.
        X : (n, c) array_like
            Covariates design matrix.
        G : (n, r) array_like
            Matrix G from the GGᵀ term.
        terms : dict
            Pre-computed terms.
        """

        self._Y = asarray(Y, float)
        self._A = asarray(A, float)
        self._X = asarray(X, float)
        self._G = asarray(G, float)
        self._H = terms["H"]
        self._logdetK = terms["logdetK"]
        self._W = terms["W"]
        self._yKiy = terms["yKiy"]
        self._WA = terms["WA"]
        self._WL0 = terms["WL0"]
        self._Lz = terms["Lz"]
        self._XRiM = terms["XRiM"]
        self._ZiXRiy = terms["ZiXRiy"]
        self._ZiXRiM = terms["ZiXRiM"]
        self._MRiM = terms["MRiM"]
        self._MRiXZiXRiM = terms["MRiXZiXRiM"]
        self._MRiy = terms["MRiy"]
        self._MRiXZiXRiy = terms["MRiXZiXRiy"]

    def null_lml(self):
        return self._null_lml

    @cached_property
    def _null_lml(self):
        """
        Log of the marginal likelihood for the null hypothesis.

        It is implemented as ::

            2·log(p(Y)) = -n·p·log(2𝜋s) - log｜K｜ - n·p,

        for which s and 𝚩 are optimal.

        Returns
        -------
        lml : float
            Log of the marginal likelihood.
        """
        np = self._nsamples * self._ntraits
        scale = self.null_scale
        return self._static_lml / 2 - np * safe_log(scale) / 2 - np / 2

    @cached_property
    def null_beta(self):
        """
        Optimal 𝛃 according to the marginal likelihood.

        It is compute by solving the equation ::

            MᵀK⁻¹M𝛃 = MᵀK⁻¹𝐲,

        for 𝐲 = vec(Y) and M = (A ⊗ X)vec(𝚩).

        Returns
        -------
        effsizes : ndarray
            Optimal 𝛃.
        """
        return rsolve(self._MKiM, self._MKiy)

    @cached_property
    def null_beta_covariance(self):
        """
        Covariance of the optimal 𝛃 according to the marginal likelihood.

        Returns
        -------
        effsizes-covariance : ndarray
            s(MᵀK⁻¹M)⁻¹.
        """
        return self.null_scale * pinv(self._H)

    @cached_property
    def null_beta_se(self):
        """
        Standard errors of the optimal 𝛃.

        Returns
        -------
        beta_se : ndarray
            Square root of the diagonal of the beta covariance.
        """
        return sqrt(self.null_beta_covariance.diagonal())

    @cached_property
    def null_scale(self):
        """
        Optimal s according to the marginal likelihood.

        The optimal s is given by

            s = (n·p)⁻¹𝐲ᵀK⁻¹(𝐲 - 𝐦),

        where 𝐦 = (A ⊗ X)vec(𝚩) and 𝚩 is optimal.

        Returns
        -------
        scale : float
            Optimal scale.
        """
        np = self._nsamples * self._ntraits
        b = vec(self.null_beta)
        mKiy = b.T @ self._MKiy
        sqrtdot = self._yKiy - mKiy
        scale = sqrtdot / np
        return scale

    def scan(self, A1, X1):
        """
        LML, fixed-effect sizes, and scale of the candidate set.

        Parameters
        ----------
        A1 : (p, e) array_like
            Trait-by-environments design matrix.
        X1 : (n, m) array_like
            Variants set matrix.

        Returns
        -------
        lml : float
            Log of the marginal likelihood for the set.
        effsizes0 : (c, p) ndarray
            Fixed-effect sizes for the covariates.
        effsizes0_se : (c, p) ndarray
            Fixed-effect size standard errors for the covariates.
        effsizes1 : (m, e) ndarray
            Fixed-effect sizes for the candidates.
        effsizes1_se : (m, e) ndarray
            Fixed-effect size standard errors for the candidates.
        scale : float
            Optimal scale.
        """
        from numpy import empty
        from numpy.linalg import multi_dot
        from numpy_sugar import epsilon, is_all_finite

        A1 = asarray(A1, float)
        X1 = asarray(X1, float)

        if not is_all_finite(A1):
            raise ValueError("A1 parameter has non-finite elements.")

        if not is_all_finite(X1):
            raise ValueError("X1 parameter has non-finite elements.")

        if A1.shape[1] == 0:
            beta_se = sqrt(self.null_beta_covariance.diagonal())
            return {
                "lml": self._null_lml,
                "effsizes0": unvec(self.null_beta, (self._ncovariates, -1)),
                "effsizes0_se": unvec(beta_se, (self._ncovariates, -1)),
                "effsizes1": empty((0,)),
                "effsizes1_se": empty((0,)),
                "scale": self.null_scale,
            }

        X1X1 = X1.T @ X1
        XX1 = self._X.T @ X1
        AWA1 = self._WA.T @ A1
        A1W = A1.T @ self._W
        GX1 = self._G.T @ X1

        MRiM1 = kron(AWA1, XX1)
        M1RiM1 = kron(A1W @ A1, X1X1)

        M1Riy = vec(multi_dot([X1.T, self._Y, A1W.T]))
        XRiM1 = kron(self._WL0.T @ A1, GX1)
        ZiXRiM1 = lu_solve(self._Lz, XRiM1)

        MRiXZiXRiM1 = self._XRiM.T @ ZiXRiM1
        M1RiXZiXRiM1 = XRiM1.T @ ZiXRiM1
        M1RiXZiXRiy = XRiM1.T @ self._ZiXRiy

        T0 = [[self._MRiM, MRiM1], [MRiM1.T, M1RiM1]]
        T1 = [[self._MRiXZiXRiM, MRiXZiXRiM1], [MRiXZiXRiM1.T, M1RiXZiXRiM1]]
        T2 = [self._MRiy, M1Riy]
        T3 = [self._MRiXZiXRiy, M1RiXZiXRiy]

        MKiM = block(T0) - block(T1)
        MKiy = block(T2) - block(T3)
        beta = rsolve(MKiM, MKiy)

        mKiy = beta.T @ MKiy
        cp = self._ntraits * self._ncovariates
        effsizes0 = unvec(beta[:cp], (self._ncovariates, self._ntraits))
        effsizes1 = unvec(beta[cp:], (X1.shape[1], A1.shape[1]))

        np = self._nsamples * self._ntraits
        sqrtdot = self._yKiy - mKiy
        scale = clip(sqrtdot / np, epsilon.tiny, inf)
        lml = self._static_lml / 2 - np * safe_log(scale) / 2 - np / 2

        effsizes_se = sqrt(clip(scale * pinv(MKiM).diagonal(), epsilon.tiny, inf))
        effsizes0_se = unvec(effsizes_se[:cp], (self._ncovariates, self._ntraits))
        effsizes1_se = unvec(effsizes_se[cp:], (X1.shape[1], A1.shape[1]))

        return {
            "lml": lml,
            "effsizes0": effsizes0,
            "effsizes1": effsizes1,
            "scale": scale,
            "effsizes0_se": effsizes0_se,
            "effsizes1_se": effsizes1_se,
        }

    @cached_property
    def _static_lml(self):
        np = self._nsamples * self._ntraits
        static_lml = -np * log2pi - self._logdetK
        return static_lml

    @property
    def _nsamples(self):
        return self._Y.shape[0]

    @property
    def _ntraits(self):
        return self._Y.shape[1]

    @property
    def _ncovariates(self):
        return self._X.shape[1]

    @cached_property
    def _MKiM(self):
        return self._MRiM - self._XRiM.T @ self._ZiXRiM

    @cached_property
    def _MKiy(self):
        return self._MRiy - self._XRiM.T @ self._ZiXRiy



from copy import deepcopy
from math import fsum

from numpy import dot, empty, inf, isfinite, log, maximum, zeros
from numpy.linalg import norm

from .._util import lu_solve
from .posterior import Posterior
from .site import Site

MAX_ITERS = 100


class EP(object):
    """
    Expectation Propagation algorithm.

    Let ::

        𝒩(𝐳｜𝐦, K)

    be the prior distribution.
    This class estimates the log of the marginal likelihood ::

        p(𝐲) = ∫∏ᵢp(yᵢ｜μᵢ = g(zᵢ))𝒩(𝐳｜𝐦, K) d𝐳

    via Expectation Propagation and provides its gradient.

    Attributes
    ----------
        _site: site-likelihood.
        _psite: previous site-likelihood.
        _cav: cavity distribution.
        _posterior: posterior distribution.
        _moments: moments for KL moment matching.
    """

    def __init__(self, nsites, posterior_type=Posterior, rtol=1.49e-05, atol=1.49e-08):
        self._site = Site(nsites)
        self._psite = Site(nsites)
        self._rtol = rtol
        self._atol = atol

        self._cav = dict(tau=zeros(nsites), eta=zeros(nsites))
        self._posterior = posterior_type(self._site)

        self._moments = {
            "log_zeroth": empty(nsites),
            "mean": empty(nsites),
            "variance": empty(nsites),
        }

        self._need_update = True
        self._cache = dict(lml=None, grad=None)
        self.verbose = False

    def reset(self):
        self._site.reset()
        self._psite.reset()

        self._cav["tau"][:] = 0
        self._cav["eta"][:] = 0
        self._posterior.reset()

        self._moments["log_zeroth"][:] = 0
        self._moments["mean"][:] = 0
        self._moments["variance"][:] = 0

        self._need_update = True
        self._cache["lml"] = None
        self._cache["grad"] = None
        self.verbose = False

    def __copy__(self):
        cls = self.__class__
        ep = cls.__new__(cls)

        ep.__dict__["_site"] = deepcopy(self.site)
        ep.__dict__["_psite"] = deepcopy(self.__dict__["_psite"])

        ep.__dict__["_rtol"] = self._rtol
        ep.__dict__["_atol"] = self._atol

        ep.__dict__["_cav"] = deepcopy(self.cav)

        posterior_type = type(self.posterior)
        ep.__dict__["_posterior"] = posterior_type(ep.site)

        ep.__dict__["_moments"] = deepcopy(self.moments)

        ep.__dict__["_need_update"] = self._need_update
        ep.__dict__["_compute_moments"] = None
        ep.__dict__["_cache"] = deepcopy(self._cache)
        ep.__dict__["verbose"] = deepcopy(self.verbose)

        return ep

    def _update(self):
        from numpy_sugar import epsilon

        if not self._need_update:
            return

        self._posterior.update()

        i = 0
        tiny = epsilon.tiny
        tol = -inf
        n1 = norm(self._site.tau)
        while i < MAX_ITERS and norm(self._site.tau - self._psite.tau) > tol:
            self._psite.tau[:] = self._site.tau
            self._psite.eta[:] = self._site.eta

            self._cav["tau"][:] = maximum(self._posterior.tau - self._site.tau, tiny)
            self._cav["eta"][:] = self._posterior.eta - self._site.eta
            self._compute_moments(self._cav["eta"], self._cav["tau"], self._moments)

            self._site.update(
                self._moments["mean"], self._moments["variance"], self._cav
            )

            self._posterior.update()

            n0 = n1
            n1 = norm(self._site.tau)
            tol = self._rtol * min(n0, n1) + self._atol
            i += 1
            if i % 10 == 9:
                self._rtol *= 10

        if i == MAX_ITERS:
            msg = "Maximum number of EP iterations has" + " been attained."
            msg += " Last EP step was: %.10f." % norm(self._site.tau - self._psite.tau)
            raise ValueError(msg)

        if self.verbose:
            print("{} EP iterations".format(i))
        self._need_update = False

    @property
    def cav(self):
        return self._cav

    def lml(self):
        from numpy_sugar.linalg import lu_slogdet

        if self._cache["lml"] is not None:
            return self._cache["lml"]

        self._update()

        LU = self._posterior.LU()
        Q, S = self._posterior.cov["QS"]
        Q = Q[0]
        ttau = self._site.tau
        teta = self._site.eta
        ctau = self._cav["tau"]
        ceta = self._cav["eta"]
        m = self._posterior.mean

        TS = ttau + ctau

        lml = [
            -0.5 * lu_slogdet(LU)[1],
            -0.5 * sum(log(S)),
            # lml += 0.5 * sum(log(ttau)),
            +0.5 * dot(teta, dot(Q, lu_solve(LU, dot(Q.T, teta)))),
            -0.5 * dot(teta, teta / TS),
            +dot(m, teta) - 0.5 * dot(m, ttau * m),
            -0.5 * dot(m * ttau, dot(Q, lu_solve(LU, dot(Q.T, 2 * teta - ttau * m)))),
            +sum(self._moments["log_zeroth"]),
            +0.5 * sum(log(TS)),
            # lml -= 0.5 * sum(log(ttau)),
            -0.5 * sum(log(ctau)),
            +0.5 * dot(ceta / TS, ttau * ceta / ctau - 2 * teta),
        ]
        lml = fsum(lml)

        if not isfinite(lml):
            raise ValueError("LML should not be %f." % lml)

        self._cache["lml"] = lml
        return lml

    def lml_derivative_over_cov(self, dQS):
        from numpy_sugar.linalg import ddot, dotd

        self._update()

        LU = self._posterior.LU()
        Q = self._posterior.cov["QS"][0][0]
        ttau = self._site.tau
        teta = self._site.eta

        diff = teta - ttau * self._posterior.mean

        v0 = dot(dQS[0][0], dQS[1] * dot(dQS[0][0].T, diff))
        v1 = ttau * dot(Q, lu_solve(LU, dot(Q.T, diff)))
        dlml = 0.5 * dot(diff, v0)
        dlml -= dot(v0, v1)
        dlml += 0.5 * dot(v1, dot(dQS[0][0], dQS[1] * dot(dQS[0][0].T, v1)))
        dqs = ddot(dQS[1], dQS[0][0].T, left=True)
        diag = dotd(dQS[0][0], dqs)
        dlml -= 0.5 * sum(ttau * diag)

        tmp = lu_solve(LU, dot(ddot(Q.T, ttau, left=False), dQS[0][0]))
        dlml += 0.5 * sum(ttau * dotd(Q, dot(tmp, dqs)))

        return dlml

    def lml_derivative_over_mean(self, dm):

        self._update()

        LU = self._posterior.LU()
        Q = self._posterior.cov["QS"][0][0]
        ttau = self._site.tau
        teta = self._site.eta

        diff = teta - ttau * self._posterior.mean

        dlml = dot(diff, dm)
        dlml -= dot(diff, dot(Q, lu_solve(LU, dot(Q.T, (ttau * dm.T).T))))

        return dlml

    @property
    def moments(self):
        return self._moments

    @property
    def posterior(self):
        return self._posterior

    def set_compute_moments(self, cm):
        self._compute_moments = cm

    @property
    def site(self):
        return self._site

    def set_prior(self, mean, covariance):
        self._posterior.mean = mean
        self._posterior.cov = covariance
        self._need_update = True
        for k in self._cache:
            self._cache[k] = None



import pytest
import scipy.stats as st
from numpy import concatenate, eye, kron
from numpy.random import default_rng
from numpy.testing import assert_allclose, assert_equal
from scipy.optimize import check_grad

from glimix_core._util import assert_interface, multivariate_normal, unvec, vec
from glimix_core.lmm import Kron2Sum


def test_kron2sum_restricted():
    random = default_rng(0)
    n = 5
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=True)

    assert_allclose(lmm.lml(), -13.9278187557)
    assert_allclose(lmm._check_grad(step=1e-7), 0, atol=1e-4)
    assert_equal(lmm.nsamples, n)
    assert_equal(lmm.ntraits, 3)
    assert_equal(lmm.ncovariates, 2)

    n = 5
    Y = random.normal(size=(n, 1))
    A = random.normal(size=(1, 1))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=True)
    lmm.name = "KronSum"

    assert_allclose(lmm.lml(), -3.8604834622)
    assert_allclose(lmm._check_grad(step=1e-7), 0, atol=1e-4)
    assert_allclose([lmm.mean()[0], lmm.mean()[1]], [-0.8736140657, 0.1044747382])

    assert_allclose(
        lmm.covariance(),
        [
            [1.4638748594, -0.5112474979, 0.4868548458, -0.608025196, -0.220970285],
            [-0.5112474979, 2.5900087221, -0.8159022113, 0.5455802776, 0.3561742975],
            [0.4868548458, -0.8159022113, 2.1094730061, -0.7091682879, 0.3513920903],
            [-0.608025196, 0.5455802776, -0.7091682879, 1.8387667433, 0.0990124051],
            [-0.220970285, 0.3561742975, 0.3513920903, 0.0990124051, 2.3558066469],
        ],
    )

    assert_equal(lmm.nsamples, n)
    assert_equal(lmm.ntraits, 1)
    assert_equal(lmm.name, "KronSum")
    lmm.fit(verbose=False)
    grad = lmm.gradient()
    assert_allclose(grad["C0.Lu"], [0], atol=1e-3)
    assert_allclose(grad["C1.Lu"], [0], atol=1e-3)
    assert_allclose(lmm.lml(), -2.2327723647, rtol=1e-5)

    A = lmm.beta_covariance
    assert_allclose(
        A,
        [[0.2173794271, -0.0067751264], [-0.0067751264, 0.163627761]],
        atol=1e-4,
        rtol=1e-4,
    )


def test_kron2sum_unrestricted():
    random = default_rng(0)
    n = 5
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)

    assert_allclose(lmm.lml(), -20.1124657252)
    assert_allclose(lmm._check_grad(step=1e-7), 0, atol=1e-4)
    assert_equal(lmm.nsamples, n)
    assert_equal(lmm.ntraits, 3)
    assert_equal(lmm.ncovariates, 2)

    n = 5
    Y = random.normal(size=(n, 1))
    A = random.normal(size=(1, 1))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lmm.name = "KronSum"

    assert_allclose(lmm.lml(), -6.400073889)
    assert_allclose(lmm._check_grad(step=1e-7), 0, atol=1e-4)
    assert_allclose(
        lmm.mean(),
        [-0.8736140657, 0.1044747382, -0.0881302921, 1.2552131472, -1.2864745361],
    )

    assert_allclose(
        lmm.covariance(),
        [
            [1.4638748594, -0.5112474979, 0.4868548458, -0.608025196, -0.220970285],
            [-0.5112474979, 2.5900087221, -0.8159022113, 0.5455802776, 0.3561742975],
            [0.4868548458, -0.8159022113, 2.1094730061, -0.7091682879, 0.3513920903],
            [-0.608025196, 0.5455802776, -0.7091682879, 1.8387667433, 0.0990124051],
            [-0.220970285, 0.3561742975, 0.3513920903, 0.0990124051, 2.3558066469],
        ],
    )

    assert_equal(lmm.nsamples, n)
    assert_equal(lmm.ntraits, 1)
    assert_equal(lmm.name, "KronSum")
    lmm.fit(verbose=False)
    grad = lmm.gradient()
    assert_allclose(grad["C0.Lu"], [0], atol=1e-3)
    assert_allclose(grad["C1.Lu"], [0], atol=1e-3)
    assert_allclose(lmm.lml(), -2.44422321522, rtol=1e-4)

    A = [
        [0.130427638, -0.0040650753],
        [-0.0040650753, 0.0981766429],
    ]
    assert_allclose(lmm.beta_covariance, A, atol=1e-4, rtol=1e-4)


def test_kron2sum_unrestricted_lml():
    random = default_rng(0)
    Y = random.normal(size=(5, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(5, 2))
    G = random.normal(size=(5, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    y = vec(lmm._Y)

    m = lmm.mean()
    K = lmm.covariance()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))

    lmm._cov.C0.Lu = random.normal(size=(3))
    m = lmm.mean()
    K = lmm.covariance()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))

    lmm._cov.C1.Lu = random.normal(size=(6))
    m = lmm.mean()
    K = lmm.covariance()
    assert_allclose(lmm.lml(), st.multivariate_normal(m, K).logpdf(y))


def test_kron2sum_public_attrs():
    callables = [
        "covariance",
        "fit",
        "get_fast_scanner",
        "gradient",
        "lml",
        "mean",
        "value",
    ]
    properties = [
        "A",
        "B",
        "C0",
        "C1",
        "M",
        "X",
        "beta",
        "beta_covariance",
        "name",
        "ncovariates",
        "nsamples",
        "ntraits",
    ]
    assert_interface(Kron2Sum, callables, properties)


def test_kron2sum_interface():
    random = default_rng(2)
    Y = random.normal(size=(2, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(2, 2))
    G = random.normal(size=(2, 4))
    with pytest.warns(UserWarning):
        lmm = Kron2Sum(Y, A, F, G, restricted=False)

    assert_allclose(
        lmm.covariance(),
        [
            [
                1.3898116696,
                -0.0921145976,
                1.3897967685,
                -0.0921145976,
                1.3897967685,
                -0.0921145976,
            ],
            [
                -0.0921145976,
                2.4263755083,
                -0.0921145976,
                2.4263606071,
                -0.0921145976,
                2.4263606071,
            ],
            [
                1.3897967685,
                -0.0921145976,
                2.3898116696,
                -0.0921145976,
                2.3897967685,
                -0.0921145976,
            ],
            [
                -0.0921145976,
                2.4263606071,
                -0.0921145976,
                3.4263755083,
                -0.0921145976,
                3.4263606071,
            ],
            [
                1.3897967685,
                -0.0921145976,
                2.3897967685,
                -0.0921145976,
                3.3898116696,
                -0.0921145976,
            ],
            [
                -0.0921145976,
                2.4263606071,
                -0.0921145976,
                3.4263606071,
                -0.0921145976,
                4.4263755083,
            ],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.mean(),
        [
            0.1890533818,
            -2.4414673826,
            -0.5227484415,
            1.7997073827,
            -0.4130635434,
            1.144165872,
        ],
        atol=1e-7,
    )
    assert_allclose(lmm.lml(), -6.1202145396, atol=1e-7)
    assert_allclose(lmm.value(), lmm.lml(), atol=1e-7)
    assert_allclose(
        lmm.A,
        [
            [0.7837560972, 0.8493429167, -0.3780225053],
            [0.8493429167, 1.3588032243, -0.7335569048],
            [-0.3780225053, -0.7335569048, 0.9426084885],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.B,
        [
            [28.569137354, -26.4006985758, -11.3407887073],
            [9.9013805057, -10.0820334998, -5.0886226006],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.X,
        [[-0.0991980517, 0.545288714], [-0.6071856999, 0.1268278471]],
        atol=1e-7,
    )
    assert_allclose(
        lmm.M,
        [
            [
                -0.0777470779,
                0.4273733543,
                -0.0842531626,
                0.4631371068,
                0.037499096,
                -0.2061314058,
            ],
            [
                -0.4758854944,
                0.0994020985,
                -0.5157088733,
                0.1077203336,
                0.2295298595,
                -0.0479437805,
            ],
            [
                -0.0842531626,
                0.4631371068,
                -0.1347906325,
                0.7409400627,
                0.0727674158,
                -0.4000003012,
            ],
            [
                -0.5157088733,
                0.1077203336,
                -0.8250458867,
                0.1723340876,
                0.4454052626,
                -0.093035443,
            ],
            [
                0.037499096,
                -0.2061314058,
                0.0727674158,
                -0.4000003012,
                -0.0935049256,
                0.5139937705,
            ],
            [
                0.2295298595,
                -0.0479437805,
                0.4454052626,
                -0.093035443,
                -0.5723383948,
                0.1195490053,
            ],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.C0,
        [
            [0.2364051393, 0.2364051393, 0.2364051393],
            [0.2364051393, 0.2364051393, 0.2364051393],
            [0.2364051393, 0.2364051393, 0.2364051393],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.C1,
        [
            [1.0000149011611938, 1.0, 1.0],
            [1.0, 2.000014901161194, 2.0],
            [1.0, 2.0, 3.000014901161194],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.beta,
        [
            28.569137354,
            9.9013805057,
            -26.4006985758,
            -10.0820334998,
            -11.3407887073,
            -5.0886226006,
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.beta_covariance,
        [
            [
                42.5707011974,
                17.6332114628,
                -46.1577363758,
                -19.8209753095,
                -27.5728424611,
                -12.3018127703,
            ],
            [
                17.6332114628,
                50.0469928942,
                -19.8209753095,
                -59.7913284342,
                -12.3018127703,
                -39.3510924252,
            ],
            [
                -46.1577363758,
                -19.8209753095,
                100.6414980719,
                40.5712408521,
                94.4461442836,
                37.2247082688,
            ],
            [
                -19.8209753095,
                -59.7913284342,
                40.5712408521,
                109.5329990419,
                37.2247082688,
                96.1051333454,
            ],
            [
                -27.5728424611,
                -12.3018127703,
                94.4461442836,
                37.2247082688,
                101.2452095823,
                38.9396030057,
            ],
            [
                -12.3018127703,
                -39.3510924252,
                37.2247082688,
                96.1051333454,
                38.9396030057,
                95.4263019998,
            ],
        ],
        atol=1e-7,
    )
    assert_equal(lmm.ncovariates, 2)
    assert_equal(lmm.nsamples, 2)
    assert_equal(lmm.ntraits, 3)
    assert_equal(lmm.name, "Kron2Sum")


def test_kron2sum_gradient_unrestricted():
    random = default_rng(5)
    Y = random.normal(size=(5, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(5, 2))
    G = random.normal(size=(5, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lmm._cov.C0.Lu = random.normal(size=3)
    lmm._cov.C1.Lu = random.normal(size=6)

    def func(x):
        lmm._cov.C0.Lu = x[:3]
        lmm._cov.C1.Lu = x[3:9]
        return lmm.lml()

    def grad(x):
        lmm._cov.C0.Lu = x[:3]
        lmm._cov.C1.Lu = x[3:9]
        D = lmm.gradient()
        return concatenate((D["C0.Lu"], D["C1.Lu"]))

    assert_allclose(
        check_grad(func, grad, random.normal(size=9), epsilon=1e-8), 0, atol=1e-3
    )


def test_kron2sum_fit_ill_conditioned_unrestricted():
    random = default_rng(0)
    n = 30
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-152.5971290143, -118.7196733895])
    grad = lmm.gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 9, atol=1e-2)


def test_kron2sum_fit_C1_well_cond_unrestricted():
    random = default_rng(0)
    Y = random.normal(size=(5, 2))
    A = random.normal(size=(2, 2))
    A = A @ A.T
    F = random.normal(size=(5, 2))
    G = random.normal(size=(5, 6))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-12.1358805747, -6.6399563682], rtol=1e-5)
    grad = lmm.gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 5, atol=1e-2)


def test_kron2sum_fit_C1_well_cond_C0_fullrank_unrestricted():
    random = default_rng(0)
    Y = random.normal(size=(5, 2))
    A = random.normal(size=(2, 2))
    A = A @ A.T
    F = random.normal(size=(5, 2))
    G = random.normal(size=(5, 6))
    with pytest.warns(UserWarning):
        lmm = Kron2Sum(Y, A, F, G, rank=2, restricted=False)
    lml0 = lmm.lml()
    lmm.fit(verbose=False)
    lml1 = lmm.lml()
    assert_allclose([lml0, lml1], [-12.7636065645, -6.639956391])
    grad = lmm.gradient()
    vars = grad.keys()
    assert_allclose(concatenate([grad[var] for var in vars]), [0] * 7, atol=1e-2)


def test_kron2sum_fit_C1_well_cond_redutant_F_unrestricted():
    random = default_rng(0)
    Y = random.normal(size=(5, 2))
    A = random.normal(size=(2, 2))
    A = A @ A.T
    F = random.normal(size=(5, 2))
    F = concatenate((F, F), axis=1)
    G = random.normal(size=(5, 2))
    with pytest.warns(UserWarning):
        Kron2Sum(Y, A, F, G, restricted=False)


def test_kron2sum_fit_C1_well_cond_redundant_Y_unrestricted():
    random = default_rng(0)
    Y = random.normal(size=(5, 2))
    Y = concatenate((Y, Y), axis=1)
    A = random.normal(size=(4, 4))
    A = A @ A.T
    F = random.normal(size=(5, 2))
    G = random.normal(size=(5, 2))
    with pytest.warns(UserWarning):
        lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lml = lmm.lml()
    assert_allclose(lml, -22.3879477239)


def test_kron2sum_large_outcome():
    random = default_rng(2)
    n = 50
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    B = random.normal(size=(2, 3))
    C0 = random.normal(size=(3, 3))
    C0 = C0 @ C0.T
    C1 = random.normal(size=(3, 3))
    C1 = C1 @ C1.T
    K = kron(C0, (G @ G.T)) + kron(C1, eye(n))
    y = multivariate_normal(random, kron(A, F) @ vec(B), K)
    Y = unvec(y, (n, 3))
    Y = Y / Y.std(0)

    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lmm.fit(verbose=False)

    assert_allclose(lmm.lml(), -71.8213415517)
    assert_allclose(lmm.C0[0, 1], -0.1672105838, rtol=1e-3, atol=1e-5)
    assert_allclose(lmm.C1[0, 1], 0.0637083841, rtol=1e-3, atol=1e-5)
    assert_allclose(lmm.beta[2], 0.0191120584, rtol=1e-3, atol=1e-5)
    assert_allclose(lmm.beta_covariance[0, 1], -0.1393152897, rtol=1e-3, atol=1e-5)
    assert_allclose(lmm.mean()[3], -0.0663303304, rtol=1e-2, atol=1e-5)
    assert_allclose(lmm.covariance()[0, 1], 0.4083172663, rtol=1e-3, atol=1e-5)


def test_kron2sum_large_covariance():
    random = default_rng(0)
    n = 50
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 4))
    scale = 1e4

    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lmm.fit(verbose=False)

    lmm_large = Kron2Sum(Y, A, F, scale * G, restricted=False)
    lmm_large.fit(verbose=False)

    assert_allclose(lmm_large.lml(), lmm.lml())
    assert_allclose(lmm_large.C0, lmm.C0 / (scale**2), rtol=1e-3, atol=1e-5)
    assert_allclose(lmm_large.C1, lmm.C1, rtol=1e-3, atol=1e-5)
    assert_allclose(lmm_large.beta, lmm.beta, rtol=1e-3, atol=1e-5)
    assert_allclose(
        lmm_large.beta_covariance, lmm.beta_covariance, rtol=1e-3, atol=1e-5
    )
    assert_allclose(lmm_large.mean(), lmm.mean(), rtol=1e-2, atol=1e-5)
    assert_allclose(lmm_large.covariance(), lmm.covariance(), rtol=1e-3, atol=1e-5)


def test_kron2sum_insufficient_sample_size():
    random = default_rng(0)
    n = 2
    Y = random.normal(size=(n, 2))
    A = random.normal(size=(2, 2))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 6))
    with pytest.warns(UserWarning):
        Kron2Sum(Y, A, F, G)



def format_function(func, params, attrs=None):
    if attrs is None:
        attrs = []
    tname = type(func).__name__
    name = func.name
    kwargs_input = [f"{arg}={val}" for arg, val in params.items()]
    input = ", ".join(kwargs_input)
    msg = f"{tname}({input})"
    if name is not None:
        msg += f": {name}"

    msg += "\n"
    for a in attrs:
        msg += _format_named_arr(a[0], a[1])
    return msg


def _format_named_arr(name, arr):
    from textwrap import TextWrapper

    prefix = f"  {name}: "
    wrapper = TextWrapper(initial_indent=prefix)
    msg = wrapper.fill(str(arr).split("\n")[0]) + "\n"
    prefix = "     "
    wrapper = TextWrapper(initial_indent=prefix, subsequent_indent=prefix)
    for s in str(arr).split("\n")[1:]:
        msg += wrapper.fill(s) + "\n"
    return msg[:-1]



from numpy import eye
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov


def test_eyecov():
    cov = EyeCov(2)
    cov.scale = 1.5
    assert_allclose(cov.value(), 1.5 * eye(2))
    assert_allclose(cov._check_grad(), 0, atol=1e-5)



"""
Generalised linear mixed models.

Fast inference for generalised linear mixed models.
"""
from . import cov, example, ggp, glmm, gp, lik, link, lmm, mean, random
from ._testit import test

__version__ = "3.1.14"

__all__ = [
    "__version__",
    "example",
    "test",
    "ggp",
    "gp",
    "lmm",
    "glmm",
    "cov",
    "lik",
    "mean",
    "link",
    "random",
]



from numpy import clip, finfo, inf, log

__all__ = ["logmax", "safe_log"]

logmax = log(finfo(float).max)


def safe_log(x):
    from numpy_sugar import epsilon

    return log(clip(x, epsilon.small, inf))



from numpy import asarray


class B:
    """
    Facilitate solving (𝑣₀𝙺 + 𝑣₁𝙸)𝐱 = 𝐲 for 𝐱, where 𝙺 is a symmetric matrix.

    Let 𝚀𝚂𝚀ᵀ = 𝙺, where 𝚀𝚂𝚀ᵀ is the eigendecomposition of 𝙺. Let 𝙳 = (𝑣₀𝚂 + 𝑣₁𝙸) and
    𝙳₀ = (𝑣₀𝚂₀ + 𝑣₁𝙸₀), where 𝚂₀ is the part of 𝚂 with positive values. Let us define:

        𝙱 = 𝚀₀𝙳₀⁻¹𝚀₀ᵀ                    if 𝑣₁=0, and
        𝙱 = 𝚀₀𝙳₀⁻¹𝚀₀ᵀ + 𝑣₁⁻¹(𝙸 - 𝚀₀𝚀₀ᵀ)  if 𝑣₁>0.

    We have 𝐱 = 𝙱𝐲.

    Parameters
    ----------
    Q0 : array_like
        𝚀₀.
    D0 : array_like
        𝙳₀.
    v0
        𝑣₀.
    v1
        𝑣₁.
    """

    def __init__(self, Q0, S0, v0: float, v1: float):

        if v0 < 0.0:
            raise ValueError("Variance `v0` must be non-negative.")

        if v1 < 0.0:
            raise ValueError("Variance `v1` must be non-negative.")

        Q0 = asarray(Q0, float)
        S0 = asarray(S0, float)
        if S0.ndim != 1:
            raise ValueError("`D0` must be an unidimensional array.")

        self._Q0 = Q0
        self._S0 = S0
        self._v0 = v0
        self._v1 = v1
        self._v0S0 = self._v0 * self._S0
        D0 = self._v0S0 + self._v1
        self._Q0D0i = self._Q0 / D0
        self._update_v0 = False
        self._update_v1 = False

    @property
    def v0(self) -> float:
        return self._v0

    @property
    def v1(self) -> float:
        return self._v1

    def set_variances(self, v0: float, v1: float):
        if v0 != self._v0:
            self._v0 = v0
            self._v1 = v1
            self._update_v0 = True
            self._update_v1 = True
        elif v1 != self._v1:
            self._v1 = v1
            self._update_v1 = True

    def dot(self, y):
        """
        Compute 𝙱𝐲.
        """
        from numpy_sugar import epsilon

        if self._update_v0:
            self._v0S0[:] = self._v0 * self._S0
            D0 = self._v0S0 + self._v1
            self._Q0D0i[:] = self._Q0 / D0
            self._update_v0 = self._update_v1 = False

        elif self._update_v1:
            D0 = self._v0S0 + self._v1
            self._Q0D0i[:] = self._Q0 / D0
            self._update_v1 = False

        Q0ty = self._Q0.T @ y
        x = self._Q0D0i @ Q0ty
        # TODO: I should check whether self._v0 is also too small
        if self._v1 > epsilon.small:
            x += (y - self._Q0 @ Q0ty) / self._v1
        return x



from numpy import empty, eye


def economic_qs_zeros(n: int):
    """
    Eigen decomposition of a zero matrix.
    """

    Q0 = empty((n, 0))
    Q1 = eye(n)
    S0 = empty(0)

    return ((Q0, Q1), S0)



from numpy import asarray, kron, zeros
from optimix import Function, Vector

from glimix_core._util import unvec, vec


class KronMean(Function):
    """
    Kronecker mean function, (A⊗X)vec(B).

    Let

    - n be the number of samples;
    - p the number of traits; and
    - c the number of covariates.

    The mathematical representation is

        𝐦 = (A⊗X)vec(B)

    where A is a p×p trait design matrix of fixed effects and X is a n×c sample design
    matrix of fixed effects. B is a c×p matrix of fixed-effect sizes.
    """

    def __init__(self, A, X):
        """
        Constructor.

        Parameters
        ----------
        A : array_like
            p×p array.
        X : array_like
            n×c array.
        """
        self._A = asarray(A, float)
        self._X = asarray(X, float)
        vecB = zeros((X.shape[1], A.shape[0])).ravel()
        self._vecB = Vector(vecB)
        self._nparams = vecB.size
        Function.__init__(self, "KronMean", vecB=self._vecB)

    @property
    def nparams(self):
        """
        Number of parameters.
        """
        return self._nparams

    @property
    def A(self):
        """
        Matrix A.
        """
        return self._A

    @property
    def X(self):
        """
        Matrix X.
        """
        return self._X

    @property
    def AX(self):
        """
        A ⊗ X.
        """
        return kron(self.A, self.X)

    def value(self):
        """
        Kronecker mean function.

        Returns
        -------
        𝐦 : ndarray
            (A⊗X)vec(B).
        """
        return self.AX @ self._vecB.value

    def gradient(self):
        """
        Gradient of the linear mean function.

        Returns
        -------
        vecB : ndarray
            Derivative of M over vec(B).
        """
        return {"vecB": self.AX}

    @property
    def B(self):
        """
        Effect-sizes parameter, B.
        """
        return unvec(self._vecB.value, (self.X.shape[1], self.A.shape[0]))

    @B.setter
    def B(self, v):
        self._vecB.value = vec(asarray(v, float))

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(A=..., X=...)".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        mat = format(self.B)
        msg += "  B: " + "\n     ".join(mat.split("\n"))
        return msg



from numpy.random import default_rng
from numpy.testing import assert_allclose

from glimix_core.cov import LinearCov


def test_linearcov():
    X = default_rng(0).normal(size=(3, 2))
    cov = LinearCov(X)
    assert_allclose(cov.value(), X @ X.T)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.scale, 1.0)
    cov.scale = 1.5
    assert_allclose(cov.scale, 1.5)
    assert_allclose(cov.value(), 1.5 * (X @ X.T))
    assert_allclose(cov._check_grad(), 0, atol=1e-5)



from numpy.random import Generator


def multivariate_normal(random: Generator, mean, cov):
    """
    Draw random samples from a multivariate normal distribution.

    Parameters
    ----------
    random
        Random state.
    mean : array_like
        Mean of the n-dimensional distribution.
    cov : array_like
        Covariance matrix of the distribution. It must be symmetric and
        positive-definite for proper sampling.

    Returns
    -------
    out : ndarray
        The drawn sample.
    """
    from numpy.linalg import cholesky

    L = cholesky(cov)
    return L @ random.normal(size=L.shape[0]) + mean



from math import fsum

from numpy import dot, isfinite, log

from .ep import EP
from .posterior_linear_kernel import PosteriorLinearKernel


def ldot(A, B):
    from numpy_sugar.linalg import ddot

    return ddot(A, B, left=True)


def dotr(A, B):
    from numpy_sugar.linalg import ddot

    return ddot(A, B, left=False)


class EPLinearKernel(EP):
    def __init__(self, nsites, rtol=1.49e-05, atol=1.49e-08):
        super(EPLinearKernel, self).__init__(
            nsites, PosteriorLinearKernel, rtol=rtol, atol=atol
        )

    def lml(self):
        from numpy_sugar.linalg import lu_slogdet

        if self._cache["lml"] is not None:
            return self._cache["lml"]

        self._update()

        LU = self._posterior.LU()
        LQt = self._posterior.LQt()
        cov = self._posterior.cov
        Q = cov["QS"][0][0]
        S = cov["QS"][1]
        ttau = self._site.tau
        teta = self._site.eta
        ctau = self._cav["tau"]
        ceta = self._cav["eta"]
        m = self._posterior.mean

        TS = ttau + ctau

        s = cov["scale"]
        d = cov["delta"]
        A = self._posterior.A
        dif = 2 * A * teta - A * ttau * m

        lml = [
            -0.5 * lu_slogdet(LU)[1],
            -0.5 * sum(log(s * S)),
            +0.5 * sum(log(A)),
            # lml += 0.5 * sum(log(ttau)),
            +0.5 * dot(teta * A, dot(Q, dot(LQt, teta * A))) * (1 - d),
            -0.5 * dot(teta, teta / TS),
            +dot(m, A * teta) - 0.5 * dot(m, A * ttau * m),
            -0.5 * dot(m * A * ttau, dot(Q, dot(LQt, dif))) * (1 - d),
            +sum(self._moments["log_zeroth"]),
            +0.5 * sum(log(TS)),
            # lml -= 0.5 * sum(log(ttau)),
            -0.5 * sum(log(ctau)),
            +0.5 * dot(ceta / TS, ttau * ceta / ctau - 2 * teta),
            0.5 * s * d * sum(teta * A * teta),
        ]
        lml = fsum(lml)

        if not isfinite(lml):
            raise ValueError("LML should not be %f." % lml)

        self._cache["lml"] = lml

        return lml

    def lml_derivatives(self, dm):
        from numpy_sugar.linalg import dotd

        if self._cache["grad"] is not None:
            return self._cache["grad"]

        self._update()

        LQt = self._posterior.LQt()
        ATQ = self._posterior.ATQ()
        ttau = self._site.tau
        teta = self._site.eta
        A = self._posterior.A

        cov = self._posterior.cov
        Q = cov["QS"][0][0]
        s = cov["scale"]
        d = cov["delta"]
        QS = self._posterior.QS()

        e_m = teta - ttau * self._posterior.mean
        Ae_m = A * e_m
        TA = ttau * A
        LtQTAe_m = dot(LQt, Ae_m)
        tQTAe_m = dot(Q.T, Ae_m)
        dKAd_m = dot(QS, tQTAe_m) * (1 - d) + d * Ae_m
        w = TA * dot(Q, LtQTAe_m) * (1 - d)
        QTAe_m = dot(Q.T, Ae_m)
        dKAs_m = -s * dot(QS, QTAe_m) + s * Ae_m

        r = (self._posterior.QSQtATQLQtA() * ttau).sum()

        dlml_mean = dot(e_m, ldot(A, dm)) - dot(
            Ae_m, dot(Q, dot(LQt, ldot(TA, dm)))
        ) * (1 - d)

        r1 = (TA * dotd(Q, LQt) * TA).sum()

        dlml_scale = 0.5 * dot(Ae_m, dKAd_m)
        dlml_scale -= sum(w * dKAd_m)
        dlml_scale += 0.5 * dot(w, dot(QS, dot(Q.T, w) * (1 - d)) + d * w)
        dlml_scale -= 0.5 * dotd(ATQ, QS.T).sum() * (1 - d)
        dlml_scale -= 0.5 * sum(TA) * d
        dlml_scale += 0.5 * r * (1 - d) * (1 - d)
        dlml_scale += 0.5 * d * r1 * (1 - d)

        dlml_delta = 0.5 * dot(Ae_m, dKAs_m)
        dlml_delta -= sum(w * dKAs_m)
        dlml_delta += 0.5 * dot(w, -s * dot(QS, dot(Q.T, w)) + s * w)
        dlml_delta += 0.5 * s * dotd(ATQ, QS.T).sum()
        dlml_delta -= 0.5 * sum(TA) * s
        dlml_delta -= 0.5 * s * r * (1 - d)
        dlml_delta += 0.5 * s * r1 * (1 - d)

        g = dict(mean=dlml_mean, scale=dlml_scale, delta=dlml_delta)

        self._cache["grad"] = g

        return g



from math import exp

from numpy import asarray, atleast_2d, dot, log, maximum, ndarray, zeros
from numpy.linalg import multi_dot, slogdet
from optimix import Function, Scalar

from .._util import (
    SVD,
    cached_property,
    economic_qs_zeros,
    log2pi,
    nice_inv,
    numbers,
    rsolve,
)
from ._b import B
from ._lmm_scan import FastScanner


class LMM(Function):
    """
    Fast Linear Mixed Models inference via maximum likelihood.

    Examples
    --------
    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from glimix_core.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.649

    One can also specify which parameters should be fitted:

    .. doctest::

        >>> from numpy import array
        >>> from numpy_sugar.linalg import economic_qs_linear
        >>> from glimix_core.lmm import LMM
        >>>
        >>> X = array([[1, 2], [3, -1]], float)
        >>> QS = economic_qs_linear(X)
        >>> covariates = array([[1], [1]])
        >>> y = array([-1, 2], float)
        >>> lmm = LMM(y, covariates, QS)
        >>> lmm.fix('delta')
        >>> lmm.fix('scale')
        >>> lmm.delta = 0.5
        >>> lmm.scale = 1
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.832
        >>> lmm.unfix('delta')
        >>> lmm.fit(verbose=False)
        >>> print('%.3f' % lmm.lml())
        -3.713

    Notes
    -----
    The LMM model can be equivalently written as ::

        𝐲 ∼ 𝓝(𝚇𝜷, 𝑠((1-𝛿)𝙺 + 𝛿𝙸)),

    and we thus have v₀ = s (1 - 𝛿) and v₁ = s 𝛿. Consider the economic
    eigendecomposition of 𝙺::

        𝙺 = [𝚀₀  𝚀₁] [𝚂₀  𝟎] [𝚀₀ᵀ]
                     [ 𝟎  𝟎] [𝚀₁ᵀ]

    and let

        𝙳 = [(1-𝛿)𝚂₀ + 𝛿𝙸₀   𝟎 ]
            [      𝟎        𝛿𝙸₁].

    In order to eliminate the need of 𝚀₁, note that 𝚀𝚀ᵀ = 𝙸 implies that ::

        𝚀₁𝚀₁ᵀ = 𝙸 - 𝚀₀𝚀₀ᵀ.

    We will need to solve ((1-𝛿)𝙺 + 𝛿𝙸)𝐱 = 𝐮 for 𝐱. Let 𝙳₀ = ((1-𝛿)𝚂₀ + 𝛿𝙸₀) and let us
    define ::

        𝙱 = 𝚀₀𝙳₀⁻¹𝚀₀ᵀ                   if 𝛿=0, and
        𝙱 = 𝚀₀𝙳₀⁻¹𝚀₀ᵀ + 𝛿⁻¹(𝙸 - 𝚀₀𝚀₀ᵀ)  if 𝛿>0.

    We therefore have ::

        𝐱 = 𝙱𝐮.
    """

    def __init__(self, y, X, QS=None, restricted=False):
        """
        Constructor.

        Parameters
        ----------
        y : array_like
            Outcome.
        X : array_like
            Covariates as a two-dimensional array.
        QS : tuple
            Economic eigendecompositon in form of ``((Q0, ), S0)`` of a
            covariance matrix ``K``.
        restricted : bool
            ``True`` for restricted maximum likelihood optimization; ``False``
            otherwise. Defaults to ``False``.
        """
        from numpy_sugar import is_all_finite

        logistic = Scalar(0.0)
        logistic.listen(self._delta_update)
        logistic.bounds = (-numbers.logmax, +numbers.logmax)
        Function.__init__(self, "LMM", logistic=logistic)
        self._logistic = logistic

        y = asarray(y, float).ravel()
        if not is_all_finite(y):
            raise ValueError("There are non-finite values in the outcome.")

        if len(y) == 0:
            raise ValueError("The outcome array is empty.")

        X = atleast_2d(asarray(X, float).T).T
        if not is_all_finite(X):
            raise ValueError("There are non-finite values in the covariates matrix.")

        self._optimal = {"beta": False, "scale": False}
        if QS is None:
            QS = economic_qs_zeros(len(y))
            self._B = B(QS[0][0], QS[1], 0.0, 1.0)
            self.delta = 1.0
            logistic.fix()
        else:
            self._B = B(QS[0][0], QS[1], 0.5, 0.5)
            self.delta = 0.5

        if QS[0][0].shape[0] != len(y):
            msg = "Sample size differs between outcome and covariance decomposition."
            raise ValueError(msg)

        if y.shape[0] != X.shape[0]:
            msg = "Sample size differs between outcome and covariates."
            raise ValueError(msg)

        self._y = y
        self._Q0 = QS[0][0]
        self._S0 = QS[1]
        self._Xsvd = SVD(X)
        self._tbeta = zeros(self._Xsvd.rank)
        self._scale = 1.0
        self._fix = {"beta": False, "scale": False}
        self._restricted = restricted

    @property
    def beta(self) -> ndarray:
        """
        Fixed-effect sizes.

        Returns
        -------
        effect-sizes
            Optimal fixed-effect sizes.

        Notes
        -----
        Setting the derivative of log(p(𝐲)) over effect sizes equal
        to zero leads to solutions 𝜷 from equation ::

            𝚇ᵀ𝙱𝚇𝜷 = 𝚇ᵀ𝙱𝐲
        """
        return rsolve(self._Xsvd.Vt, rsolve(self._Xsvd.US, self.mean()))

    @beta.setter
    def beta(self, beta):
        beta = asarray(beta, float).ravel()
        self._tbeta[:] = self._Xsvd.Vt @ beta
        self._optimal["beta"] = False
        self._optimal["scale"] = False

    @property
    def beta_covariance(self):
        """
        Estimates the covariance-matrix of the optimal beta.

        Returns
        -------
        beta-covariance : ndarray
            𝑠(𝚇ᵀ𝙱𝚇)⁻¹.

        References
        ----------
        .. Rencher, A. C., & Schaalje, G. B. (2008). Linear models in statistics. John
           Wiley & Sons.
        """
        A = nice_inv(self._tXTBtX)
        VT = self._Xsvd.Vt
        H = rsolve(VT, A)
        return rsolve(VT, H.T) * self.scale

    def fix(self, param: str):
        """
        Disable parameter optimization.

        Parameters
        ----------
        param
            Possible values are ``"delta"``, ``"beta"``, and ``"scale"``.
        """
        if param == "delta":
            super()._fix("logistic")
        else:
            self._fix[param] = True

    def unfix(self, param: str):
        """
        Enable parameter optimization.

        Parameters
        ----------
        param
            Possible values are ``"delta"``, ``"beta"``, and ``"scale"``.
        """
        if param == "delta":
            self._unfix("logistic")
        else:
            self._fix[param] = False

    @property
    def v0(self):
        """
        First variance.

        Returns
        -------
        v0 : float
            s(1 - 𝛿).
        """
        return self.scale * (1 - self.delta)

    @property
    def v1(self) -> float:
        """
        Second variance.

        Returns
        -------
        v1
            s𝛿.
        """
        return self.scale * self.delta

    def fit(self, verbose=True):
        """
        Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool, optional
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        """
        if not self._isfixed("logistic"):
            self._maximize_scalar(desc="LMM", rtol=1e-6, atol=1e-6, verbose=verbose)

        if not self._fix["beta"]:
            self._update_beta()

        if not self._fix["scale"]:
            self._update_scale()

    def get_fast_scanner(self) -> FastScanner:
        """
        Return :class:`.FastScanner` for association scan.

        Returns
        -------
        fast-scanner
            Instance of a class designed to perform very fast association scan.
        """
        v0 = self.v0
        v1 = self.v1
        QS = ((self._Q0,), v0 * self._S0)
        return FastScanner(self._y, self.X, QS, v1)

    def value(self) -> float:
        """
        Internal use only.
        """
        if not self._fix["beta"]:
            self._update_beta()

        if not self._fix["scale"]:
            self._update_scale()

        return self.lml()

    def gradient(self):
        """
        Not implemented.
        """
        raise NotImplementedError

    @property
    def nsamples(self) -> int:
        """
        Number of samples.
        """
        return len(self._y)

    @property
    def ncovariates(self) -> int:
        """
        Number of covariates.
        """
        return self._Xsvd.A.shape[1]

    def lml(self) -> float:
        """
        Log of the marginal likelihood.

        Returns
        -------
        lml
            Log of the marginal likelihood.

        Notes
        -----
        The log of the marginal likelihood is given by ::

            2⋅log(p(𝐲)) = -n⋅log(2π) - n⋅log(s) - log|𝙳|
                -  (𝚀₀ᵀ𝐲)ᵀ(s𝙳₀)⁻¹(𝚀₀ᵀ𝐲)  -  (𝐲)ᵀ(s𝛿)⁻¹(𝐲)  +  (𝚀₀ᵀ𝐲)ᵀ(s𝛿)⁻¹(𝚀₀ᵀ𝐲)
                - (𝚀₀ᵀ𝚇𝜷)ᵀ(s𝙳₀)⁻¹(𝚀₀ᵀ𝚇𝜷) - (𝚇𝜷)ᵀ(s𝛿)⁻¹(𝚇𝜷) + (𝚀₀ᵀ𝚇𝜷)ᵀ(s𝛿)⁻¹(𝚀₀ᵀ𝚇𝜷)
                + 2(𝚀₀ᵀ𝐲)ᵀ(s𝙳₀)⁻¹(𝚇𝜷)    + 2(𝐲)ᵀ(s𝛿)⁻¹(𝚇𝜷) - 2(𝚀₀ᵀ𝐲)ᵀ(s𝛿)⁻¹(𝚀₀ᵀ𝚇𝜷)

        By using the optimal 𝜷, the log of the marginal likelihood can be rewritten
        as::

            2⋅log(p(𝐲)) = -n⋅log(2π) - n⋅log(s) - log|𝙳| + (𝚀₀ᵀ𝐲)ᵀ(s𝙳₀)⁻¹𝚀₀ᵀ(𝚇𝜷 - 𝐲)
                        + (𝐲)ᵀ(s𝛿)⁻¹(𝚇𝜷 - 𝐲) - (𝚀₀ᵀ𝐲)ᵀ(s𝛿)⁻¹𝚀₀ᵀ(𝚇𝜷 - 𝐲).

        In the extreme case where 𝜷 is such that 𝐲 = 𝚇𝜷, the maximum is attained as
        s→0.

        For optimals 𝜷 and s, the log of the marginal likelihood can be further
        simplified to ::

            2⋅log(p(𝐲; 𝜷, s)) = -n⋅log(2π) - n⋅log s - log|𝙳| - n.
        """
        reml = (self._logdetXX - self._logdetH()) / 2
        if self._optimal["scale"]:
            lml = self._lml_optimal_scale()
        else:
            lml = self._lml_arbitrary_scale()
        return lml + reml

    @property
    def X(self):
        """
        Covariates matrix.

        Returns
        -------
        X : ndarray
            Covariates.
        """
        return self._Xsvd.A

    @property
    def delta(self) -> float:
        """
        Variance ratio between ``K`` and ``I``.
        """
        from numpy_sugar import epsilon

        v = float(self._logistic.value)

        if v > 0.0:
            v = 1 / (1 + exp(-v))
        else:
            v = exp(v)
            v = v / (v + 1.0)

        return min(max(v, epsilon.tiny), 1 - epsilon.tiny)

    @delta.setter
    def delta(self, delta):
        from numpy_sugar import epsilon

        delta = min(max(delta, epsilon.tiny), 1 - epsilon.tiny)
        self._logistic.value = log(delta / (1 - delta))
        self._optimal["beta"] = False
        self._optimal["scale"] = False

    @property
    def scale(self) -> float:
        """
        Scaling factor.

        Returns
        -------
        scale : float
            Scaling factor.

        Notes
        -----
        Setting the derivative of log(p(𝐲; 𝜷)), for which 𝜷 is optimal, over
        scale equal to zero leads to the maximum ::

            s = n⁻¹(Qᵀ𝐲)ᵀD⁻¹ Qᵀ(𝐲-𝚇𝜷).

        In the case of restricted marginal likelihood ::

            s = (n-c)⁻¹(Qᵀ𝐲)ᵀD⁻¹ Qᵀ(𝐲-𝚇𝜷),

        where s is the number of covariates.
        """
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = scale
        self._optimal["scale"] = False

    def mean(self):
        """
        Mean of the prior.

        Formally, 𝐦 = 𝚇𝜷.

        Returns
        -------
        mean : ndarray
            Mean of the prior.
        """
        return self._Xsvd.US @ self._tbeta

    def covariance(self):
        """
        Covariance of the prior.

        Returns
        -------
        covariance : ndarray
            v₀𝙺 + v₁𝙸.
        """
        from numpy_sugar.linalg import ddot, sum2diag

        Q0 = self._Q0
        S0 = self._S0
        return sum2diag(dot(ddot(Q0, self.v0 * S0), Q0.T), self.v1)

    def _delta_update(self):
        self._optimal["beta"] = False
        self._optimal["scale"] = False
        delta = self.delta
        self._B.set_variances(1 - delta, delta)

    @cached_property
    def _logdetXX(self):
        """
        log(｜XᵀX｜).
        """
        if not self._restricted:
            return 0.0

        ldet = slogdet(self._Xsvd.US.T @ self._Xsvd.US)
        if ldet[0] != 1.0:
            raise ValueError("The determinant of XᵀX should be positive.")
        return ldet[1]

    def _logdetH(self):
        """
        log(｜H｜) for H = s⁻¹XᵀQD⁻¹QᵀX.
        """
        if not self._restricted:
            return 0.0
        ldet = slogdet(self._tXTBtX / self.scale)
        if ldet[0] != 1.0:
            raise ValueError("The determinant of H should be positive.")
        return ldet[1]

    def _lml_optimal_scale(self):
        """
        Log of the marginal likelihood for optimal scale.

        Implementation for unrestricted LML::

        Returns
        -------
        lml : float
            Log of the marginal likelihood.
        """
        assert self._optimal["scale"]

        n = len(self._y)
        lml = -self._df * log2pi - self._df - n * log(self.scale)
        lml -= self._logdetD
        return lml / 2

    def _lml_arbitrary_scale(self):
        """
        Log of the marginal likelihood for arbitrary scale.

        Returns
        -------
        lml : float
            Log of the marginal likelihood.
        """
        s = self.scale
        n = len(self._y)
        lml = -self._df * log2pi - n * log(s)
        lml -= self._logdetD
        my = self.mean() - self._y
        lml -= my.T @ self._B.dot(my) / s
        return lml / 2

    @property
    def _df(self):
        """
        Degrees of freedom.
        """
        if not self._restricted:
            return self.nsamples
        return self.nsamples - self._Xsvd.rank

    def _optimal_scale_using_optimal_beta(self):
        from numpy_sugar import epsilon

        assert self._optimal["beta"]
        s = self._yTBy - self._yTBtX @ self._tbeta
        return maximum(s / self._df, epsilon.small)

    def _update_beta(self):
        assert not self._fix["beta"]
        if self._optimal["beta"]:
            return

        self._tbeta[:] = rsolve(self._tXTBtX, self._yTBtX)
        self._optimal["beta"] = True
        self._optimal["scale"] = False

    def _update_scale(self):
        from numpy_sugar import epsilon

        if self._optimal["beta"]:
            self._scale = self._optimal_scale_using_optimal_beta()
        else:
            p0 = self._yTBy - 2 * self._yTBtX @ self._tbeta
            p1 = multi_dot((self._tbeta, self._tXTBtX, self._tbeta))
            self._scale = maximum((p0 + p1) / self._df, epsilon.small)

        self._optimal["scale"] = True

    @property
    def _logdetD(self):
        v = 0.0
        d = self.delta
        rank = self._S0.size
        if rank > 0:
            v += log((1 - d) * self._S0 + d).sum()
        rankdef = self._y.shape[0] - rank
        v += rankdef * log(d)
        return v

    @property
    def _yTBy(self):
        return self._y.T @ self._B.dot(self._y)

    @property
    def _yTBtX(self):
        return self._y.T @ self._B.dot(self._Xsvd.US)

    @property
    def _tXTBtX(self):
        return self._Xsvd.US.T @ self._B.dot(self._Xsvd.US)



"""Product of likelihood functions.

BernoulliProdLik  Bernoulli likelihood.
BinomialProdLik   Binomial likelihood.
DeltaProdLik      Delta likelihood.
PoissonProdLik    Poisson likelihood.
"""
from ._prod import BernoulliProdLik, BinomialProdLik, DeltaProdLik, PoissonProdLik

__all__ = ["DeltaProdLik", "BernoulliProdLik", "BinomialProdLik", "PoissonProdLik"]



import pytest
from numpy import (
    arange,
    asarray,
    ascontiguousarray,
    corrcoef,
    dot,
    exp,
    eye,
    ones,
    sqrt,
    stack,
    zeros,
)
from numpy.random import default_rng
from numpy.testing import assert_, assert_allclose
from numpy_sugar.linalg import economic_qs, economic_qs_linear
from glimix_core._util import multivariate_normal

from glimix_core.example import linear_eye_cov
from glimix_core.glmm import GLMMExpFam, GLMMNormal
from glimix_core.random import bernoulli_sample

ATOL = 1e-3
RTOL = 1e-3


def test_glmmexpfam_layout():
    y = asarray([1.0, 0.5])
    X = asarray([[0.5, 1.0]])
    K = asarray([[1.0, 0.0], [0.0, 1.0]])
    QS = economic_qs(K)

    with pytest.raises(ValueError):
        GLMMExpFam(y, "poisson", X, QS=QS)

    y = asarray([1.0])
    with pytest.raises(ValueError):
        GLMMExpFam(y, "poisson", X, QS=QS)


def test_glmmexpfam_copy():
    nsamples = 10

    random = default_rng(2)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    z = multivariate_normal(random, 0.2 * ones(nsamples), K)
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for i, ni in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.random(size=ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm0 = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)

    assert_allclose(glmm0.lml(), -28.7031933618, atol=ATOL, rtol=RTOL)
    glmm0.fit(verbose=False)

    v = -7.0665573503
    assert_allclose(glmm0.lml(), v)

    glmm1 = glmm0.copy()
    assert_allclose(glmm1.lml(), v, rtol=1e-3)

    glmm1.scale = 0.92
    assert_allclose(glmm0.lml(), v, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm1.lml(), -195.013480839, atol=ATOL, rtol=RTOL)

    glmm0.fit(verbose=False)
    glmm1.fit(verbose=False)

    v = -7.0665573503
    assert_allclose(glmm0.lml(), v)
    assert_allclose(glmm1.lml(), v, atol=ATOL)


def test_glmmexpfam_precise():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = [random.integers(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ["binomial", ntri], X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.scale = 1.0
    assert_allclose(glmm.lml(), -29.6908906579, atol=ATOL, rtol=RTOL)
    glmm.scale = 2.0
    assert_allclose(glmm.lml(), -27.6417735686, atol=ATOL, rtol=RTOL)
    glmm.scale = 3.0
    assert_allclose(glmm.lml(), -27.2379649267, atol=ATOL, rtol=RTOL)
    glmm.scale = 4.0
    assert_allclose(glmm.lml(), -27.243363099, atol=ATOL, rtol=RTOL)
    glmm.scale = 5.0
    assert_allclose(glmm.lml(), -27.3896549751, atol=ATOL, rtol=RTOL)
    glmm.scale = 6.0
    assert_allclose(glmm.lml(), -27.5897920258, atol=ATOL, rtol=RTOL)
    glmm.delta = 0.1
    assert_allclose(glmm.lml(), -28.032630672, atol=ATOL, rtol=RTOL)

    assert_allclose(glmm._check_grad(), 0, atol=1e-3, rtol=RTOL)


def test_glmmexpfam_glmmnormal_get_fast_scanner():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    eta = random.random(size=nsamples)
    tau = 10 * random.uniform(size=nsamples)

    glmm = GLMMNormal(eta, tau, X, QS)
    glmm.fit(verbose=False)
    want = [
        0.176704059082,
        -0.248669792939,
        0.165562054326,
        -0.068362796775,
        0.160376987127,
    ]
    assert_allclose(glmm.beta, want, atol=1e-3, rtol=1e-3)
    assert_allclose(0.001, glmm.scale, atol=1e-3, rtol=1e-3)
    assert_allclose(0.999999994119, glmm.delta, atol=1e-3, rtol=1e-3)

    scanner = glmm.get_fast_scanner()
    r = scanner.fast_scan(X, verbose=False)

    assert_allclose(
        r["lml"],
        [
            6.251971978275,
            6.251971978275,
            6.251971978275,
            6.251971978275,
            6.251971978275,
        ],
        rtol=1e-6,
    )
    assert_allclose(
        r["effsizes0"],
        [
            [0.0883515605, -0.2486713745, 0.1655632824, -0.0683615838, 0.1603763469],
            [0.176703121, -0.1243356873, 0.1655632824, -0.0683615838, 0.1603763469],
            [0.176703121, -0.2486713745, 0.0827816412, -0.0683615838, 0.1603763469],
            [0.176703121, -0.2486713745, 0.1655632824, -0.0341807919, 0.1603763469],
            [0.176703121, -0.2486713745, 0.1655632824, -0.0683615838, 0.0801881735],
        ],
        rtol=1e-6,
    )
    assert_allclose(
        r["effsizes1"],
        [0.0883515605, -0.1243356873, 0.0827816412, -0.0341807919, 0.0801881735],
        rtol=1e-6,
    )
    assert_allclose(
        r["scale"],
        [0.0503128779, 0.0503128779, 0.0503128779, 0.0503128779, 0.0503128779],
        rtol=1e-6,
    )


def test_glmmexpfam_delta0():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = [random.integers(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.delta = 0

    assert_allclose(glmm.lml(), -29.2156146315, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm._check_grad(step=2e-5), 0, atol=1e-2)


def test_glmmexpfam_delta1():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = [random.integers(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.delta = 1

    assert_allclose(glmm.lml(), -31.014766648, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm._check_grad(), 0, atol=1e-4)


def test_glmmexpfam_wrong_qs():
    random = default_rng(0)
    X = random.random(size=(10, 15))
    linear_eye_cov().value()
    QS = [0, 1]

    ntri = random.integers(1, 30, 10)
    nsuc = [random.integers(0, i) for i in ntri]

    with pytest.raises(ValueError):
        GLMMExpFam((nsuc, ntri), "binomial", X, QS)


def test_glmmexpfam_optimize():
    nsamples = 10

    random = default_rng(1)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    z = multivariate_normal(random, 0.2 * ones(nsamples), K)
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for i, ni in enumerate(ntri):
        x = random.random(size=ni)
        nsuc[i] += sum(z[i] + 0.2 * x > 0)

    ntri = ascontiguousarray(ntri)
    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)

    assert_allclose(glmm.lml(), -35.9777836472, atol=ATOL, rtol=RTOL)
    glmm.fix("beta")
    glmm.fix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -32.6840270428, atol=ATOL, rtol=RTOL)

    glmm.unfix("beta")
    glmm.unfix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -1.0447594303e-05, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_optimize_low_rank():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = dot(X, X.T)
    z = dot(X, 0.2 * random.random(size=5))
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for i, ni in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.random(size=ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)

    assert_allclose(glmm.lml(), -16.1748642332, atol=ATOL, rtol=RTOL)
    glmm.fit(verbose=False)
    assert_allclose(glmm.lml(), -7.7868635894e-08, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_bernoulli_problematic():
    random = default_rng(1)
    N = 30
    G = random.random(size=(N, N + 50))
    y = bernoulli_sample(0.0, G, random_state=random)

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y), 1))
    model = GLMMExpFam(y, "bernoulli", X, QS=(QS[0], QS[1]))
    model.delta = 0
    model.fix("delta")
    model.fit(verbose=False)
    assert_allclose(model.lml(), -18.9974167099, atol=ATOL, rtol=RTOL)
    assert_allclose(model.delta, 0, atol=1e-3)
    assert_allclose(model.scale, 7.6792483084, atol=ATOL, rtol=RTOL)
    assert_allclose(model.beta, [1.0464641056], atol=ATOL, rtol=RTOL)


def test_glmmexpfam_bernoulli_probit_problematic():
    random = default_rng(1)
    N = 30
    G = random.random(size=(N, N + 50))
    y = bernoulli_sample(0.0, G, random_state=random)

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y), 1))
    model = GLMMExpFam(y, "probit", X, QS=(QS[0], QS[1]))
    model.delta = 0
    model.fix("delta")
    model.fit(verbose=False)
    assert_allclose(model.lml(), -19.0071728127, atol=ATOL, rtol=RTOL)
    assert_allclose(model.delta, 1.4901161194e-08, atol=1e-3)
    assert_allclose(model.scale, 2.6268716097, atol=ATOL, rtol=RTOL)
    assert_allclose(model.beta, [0.6039762584], atol=ATOL, rtol=RTOL)

    h20 = model.scale * (1 - model.delta) / (model.scale + 1)

    model.unfix("delta")
    model.delta = 0.5
    model.scale = 1.0
    model.fit(verbose=False)

    assert_allclose(model.lml(), -19.0071728127, atol=ATOL, rtol=RTOL)
    assert_allclose(model.delta, 0.1310042942, atol=1e-3)
    assert_allclose(model.scale, 5.0048839203, atol=ATOL, rtol=RTOL)
    assert_allclose(model.beta, [0.7771521303], atol=ATOL, rtol=RTOL)

    h21 = model.scale * (1 - model.delta) / (model.scale + 1)

    assert_allclose(h20, h21, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_bernoulli_probit_assure_delta_fixed():
    random = default_rng(1)
    N = 10
    G = random.random(size=(N, N + 50))
    y = bernoulli_sample(0.0, G, random_state=random)

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y), 1))
    model = GLMMExpFam(y, "probit", X, QS=(QS[0], QS[1]))
    model.fit(verbose=False)

    assert_allclose(model.lml(), -6.7303089368, rtol=RTOL)
    assert_allclose(model.delta, -1.7828204051e-06, atol=1e-5)
    assert_(model._isfixed("logitdelta"))


def _stdnorm(X, axis=None, out=None):
    X = ascontiguousarray(X)
    if out is None:
        out = X.copy()

    m = out.mean(axis)
    s = out.std(axis)
    ok = s > 0

    out -= m

    if out.ndim == 1:
        if s > 0:
            out /= s
    else:
        out[..., ok] /= s[ok]

    return out


def test_glmmexpfam_binomial_pheno_list():
    random = default_rng(0)
    n = 10

    X = random.random(size=(n, 2))
    G = random.random(size=(n, 100))
    K = dot(G, G.T)
    ntrials = random.integers(1, 100, n)
    z = dot(G, random.random(size=100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i in range(len(ntrials)):
        for _ in range(ntrials[i]):
            successes[i] += int(z[i] + 0.1 * random.random() > 0)

    QS = economic_qs(K)
    glmm = GLMMExpFam(successes, ("binomial", ntrials), X, QS)
    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -1.7828204051e-06, rtol=RTOL, atol=ATOL)


def test_glmmexpfam_binomial_large_ntrials():
    random = default_rng(0)
    n = 10

    X = random.random(size=(n, 2))
    G = random.random(size=(n, 100))
    K = dot(G, G.T)
    ntrials = random.integers(1, 100000, n)
    z = dot(G, random.random(size=100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i in range(len(ntrials)):
        for _ in range(ntrials[i]):
            successes[i] += int(z[i] + 0.1 * random.random() > 0)

    QS = economic_qs(K)
    glmm = GLMMExpFam(successes, ("binomial", ntrials), X, QS)
    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -8.2741657696e-05, rtol=RTOL, atol=ATOL)


def test_glmmexpfam_scale_very_low():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = [random.integers(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.scale = 1e-3
    assert_allclose(glmm.lml(), -64.4994083456, atol=ATOL, rtol=RTOL)

    assert_allclose(glmm._check_grad(), 0, atol=1e-2)


def test_glmmexpfam_scale_very_high():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    QS = economic_qs(K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = [random.integers(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4])

    glmm.scale = 30.0
    assert_allclose(glmm.lml(), -31.4476266883, atol=ATOL, rtol=RTOL)

    assert_allclose(glmm._check_grad(), 0, atol=1e-3)


def test_glmmexpfam_delta_one_zero():
    random = default_rng(1)
    n = 30
    X = random.random(size=(n, 6))
    K = dot(X, X.T)
    K /= K.diagonal().mean()
    QS = economic_qs(K)

    ntri = random.integers(1, 30, n)
    nsuc = [random.integers(0, i) for i in ntri]

    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, QS)
    glmm.beta = asarray([1.0, 0, 0.5, 0.1, 0.4, -0.2])

    glmm.delta = 0
    assert_allclose(glmm.lml(), -104.5701519999)
    assert_allclose(glmm._check_grad(step=1e-4), 0, atol=1e-2)

    glmm.fit(verbose=False)
    assert_allclose(glmm.lml(), -91.7868283938, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm.delta, 0, atol=ATOL, rtol=RTOL)

    glmm.delta = 1
    assert_allclose(glmm.lml(), -91.5788328052, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm._check_grad(step=1e-4), 0, atol=1e-1)

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -71.1968560204, atol=ATOL, rtol=RTOL)
    assert_allclose(glmm.delta, 0.9999999850988439, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_predict():
    random = default_rng(1)
    n = 100
    p = n + 1

    X = ones((n, 2))
    X[:, 1] = random.random(size=n)

    G = random.random(size=(n, p))
    G /= G.std(0)
    G -= G.mean(0)
    G /= sqrt(p)
    K = dot(G, G.T)

    i = asarray(arange(0, n), int)
    si = random.choice(i, n, replace=False)
    ntest = int(n // 5)
    itrain = si[:-ntest]
    itest = si[-ntest:]

    Xtrain = X[itrain, :]
    Ktrain = K[itrain, :][:, itrain]

    Xtest = X[itest, :]

    beta = random.random(size=2)
    z = multivariate_normal(random, dot(X, beta), 0.9 * K + 0.1 * eye(n))

    ntri = random.integers(1, 100, n)
    nsuc = zeros(n, dtype=int)
    for i, ni in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.random(size=ni) > 0)

    ntri = ascontiguousarray(ntri)

    QStrain = economic_qs(Ktrain)
    nsuc_train = ascontiguousarray(nsuc[itrain])
    ntri_train = ascontiguousarray(ntri[itrain])

    nsuc_test = ascontiguousarray(nsuc[itest])
    ntri_test = ascontiguousarray(ntri[itest])

    glmm = GLMMExpFam(nsuc_train, ("binomial", ntri_train), Xtrain, QStrain)
    glmm.fit(verbose=False)
    ks = K[itest, :][:, itrain]
    kss = asarray([K[i, i] for i in itest])
    pm = glmm.predictive_mean(Xtest, ks, kss)
    pk = glmm.predictive_covariance(Xtest, ks, kss)
    r = nsuc_test / ntri_test
    assert_(corrcoef([pm, r])[0, 1] > 0.5)
    assert_allclose(pk[0], 320.4925540377, rtol=1e-6)


def test_glmmexpfam_qs_none():
    nsamples = 10

    random = default_rng(0)
    X = random.random(size=(nsamples, 5))
    K = linear_eye_cov().value()
    z = multivariate_normal(random, 0.2 * ones(nsamples), K)

    ntri = random.integers(1, 30, nsamples)
    nsuc = zeros(nsamples, dtype=int)
    for i, ni in enumerate(ntri):
        nsuc[i] += sum(z[i] + 0.2 * random.random(size=ni) > 0)

    ntri = ascontiguousarray(ntri)
    glmm = GLMMExpFam(nsuc, ("binomial", ntri), X, None)

    assert_allclose(glmm.lml(), -56.5137400209, atol=ATOL, rtol=RTOL)
    glmm.fix("beta")
    glmm.fix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -43.4232487385, atol=ATOL, rtol=RTOL)

    glmm.unfix("beta")
    glmm.unfix("scale")

    glmm.fit(verbose=False)

    assert_allclose(glmm.lml(), -6.745656177, atol=ATOL, rtol=RTOL)


def test_glmmexpfam_poisson():
    random = default_rng(1)

    # sample size
    n = 30

    # covariates
    offset = ones(n) * random.random()
    age = random.integers(16, 75, n)
    M = stack((offset, age), axis=1)

    # genetic variants
    G = random.random(size=(n, 4))

    # sampling the phenotype
    alpha = random.random(size=2)
    beta = random.random(size=4)
    eps = random.random(size=n)
    y = M @ alpha + G @ beta + eps

    # Whole genotype of each sample.
    X = random.random(size=(n, 50))
    # Estimate a kinship relationship between samples.
    X_ = (X - X.mean(0)) / X.std(0) / sqrt(X.shape[1])
    K = X_ @ X_.T + eye(n) * 0.1
    # Update the phenotype
    y += multivariate_normal(random, zeros(n), K)
    y = (y - y.mean()) / y.std()

    z = y.copy()
    y = random.poisson(exp(z))

    M = M - M.mean(0)
    QS = economic_qs(K)
    glmm = GLMMExpFam(y, "poisson", M, QS)
    assert_allclose(glmm.lml(), -48.8750807107)
    glmm.fit(verbose=False)
    assert_allclose(glmm.lml(), -38.7780148893)



import pytest
import scipy.stats as st
from numpy import concatenate, exp, eye, log, pi
from numpy.linalg import slogdet, solve
from numpy.random import Generator, default_rng
from numpy.testing import assert_allclose, assert_equal
from numpy_sugar.linalg import ddot, economic_qs_linear, economic_svd
from scipy.optimize import minimize

from glimix_core._util import assert_interface
from glimix_core.lmm import LMM


def test_lmm():
    random = default_rng(12)

    (y, X, G) = _full_rank(random)
    _test_lmm(random, y, X, G, _get_mvn(y, X, G), False)

    (y, X, G) = _low_rank(random)
    _test_lmm(random, y, X, G, _get_mvn(y, X, G), False)

    (y, X, G) = _full_rank(random)
    _test_lmm(random, y, X, G, _get_mvn_restricted(y, X, G), True)

    (y, X, G) = _low_rank(random)
    _test_lmm(random, y, X, G, _get_mvn_restricted(y, X, G), True)


def test_lm():
    random = default_rng(0)

    (y, X, _) = _full_rank(random)
    lmm = LMM(y, X)
    lmm.fit(verbose=False)
    assert_allclose(lmm.v0, 2.0129061033356781e-16, atol=1e-7)
    assert_allclose(lmm.v1, 0.8108150965)
    assert_allclose(lmm.beta, [0.3225896208, -0.2917152789])


def test_lmm_beta_covariance():
    random = default_rng(0)

    (y, X, G) = _full_rank(random)
    QS = economic_qs_linear(G)
    lmm = LMM(y, X, QS)
    lmm.fit(verbose=False)

    A = [[0.0016815025, -0.0001486849], [-0.0001486849, 0.0023484078]]
    assert_allclose(lmm.beta_covariance, A, rtol=1e-6)

    (y, X, G) = _low_rank(random)
    QS = economic_qs_linear(G)
    lmm = LMM(y, X[:, :2], QS)
    lmm.fit(verbose=False)

    A = [[0.0037514935, -0.00114132], [-0.00114132, 0.006491444]]
    assert_allclose(lmm.beta_covariance, A)

    (y, X, G) = _low_rank(random)
    QS = economic_qs_linear(G)
    lmm = LMM(y, X, QS)
    lmm.fit(verbose=False)

    A = [
        [0.0042288753, 0.0016227908, 0.0042288753, 0.0016227908],
        [0.0016227908, 0.0080978309, 0.0016227908, 0.0080978309],
        [0.0042288753, 0.0016227908, 0.0042288753, 0.0016227908],
        [0.0016227908, 0.0080978309, 0.0016227908, 0.0080978309],
    ]
    assert_allclose(lmm.beta_covariance, A)


def test_lmm_public_attrs():
    callables = [
        "covariance",
        "fit",
        "fix",
        "get_fast_scanner",
        "gradient",
        "lml",
        "mean",
        "unfix",
        "value",
    ]
    properties = [
        "X",
        "beta",
        "beta_covariance",
        "delta",
        "name",
        "ncovariates",
        "nsamples",
        "scale",
        "v0",
        "v1",
    ]
    assert_interface(LMM, callables, properties)


def test_lmm_interface():
    random = default_rng(1)
    n = 3
    G = random.normal(size=(n, n + 1))
    X = random.normal(size=(n, 2))
    y = (
        X @ random.normal(size=2)
        + G @ random.normal(size=G.shape[1])
        + random.normal(size=n)
    )
    y -= y.mean(0)
    y /= y.std(0)

    QS = economic_qs_linear(G)
    lmm = LMM(y, X, QS, restricted=False)
    lmm.name = "lmm"
    lmm.fit(verbose=False)

    assert_allclose(
        lmm.covariance(),
        [
            [1.0454767788, -0.1024992983, -0.1347709001],
            [-0.1024992983, 0.6609585597, 0.3069073261],
            [-0.1347709001, 0.3069073261, 0.2085936602],
        ],
        atol=1e-7,
    )
    assert_allclose(
        lmm.mean(),
        [-0.259134127, -1.4618988298, 0.5657654358],
        atol=1e-7,
    )
    assert_allclose(lmm.lml(), -2.6467351496, atol=1e-7)
    assert_allclose(lmm.value(), lmm.lml(), atol=1e-7)
    assert_allclose(lmm.lml(), -2.6467351496, atol=1e-7)
    assert_allclose(
        lmm.X,
        [
            [-0.736454087, -0.162909948],
            [-0.4821193127, 0.5988462126],
            [0.0397221075, -0.292456751],
        ],
        atol=1e-7,
    )
    assert_allclose(lmm.beta, [0.757055469, -1.8317019366], atol=1e-7)
    assert_allclose(
        lmm.beta_covariance,
        [[1.2672251036, 0.4685773896], [0.4685773896, 0.3343485385]],
        atol=1e-7,
    )
    assert_allclose(lmm.delta, 2.2204460493e-16, atol=1e-7)
    assert_equal(lmm.ncovariates, 2)
    assert_equal(lmm.nsamples, 3)
    assert_allclose(lmm.scale, 0.4018140195, atol=1e-7)
    assert_allclose(lmm.v0, 0.4018140195, atol=1e-7)
    assert_allclose(lmm.v1, 8.9220635204e-17, atol=1e-7)
    assert_equal(lmm.name, "lmm")

    with pytest.raises(NotImplementedError):
        lmm.gradient()


def _test_lmm(random: Generator, y, X, G, mvn, restricted):
    c = X.shape[1]
    QS = economic_qs_linear(G)
    lmm = LMM(y, X, QS, restricted=restricted)
    beta = lmm.beta
    v0 = lmm.v0
    v1 = lmm.v1

    K0 = G @ G.T
    assert_allclose(lmm.lml(), mvn(beta, v0, v1, y, X, K0))

    beta = random.normal(size=c)
    lmm.beta = beta
    assert_allclose(lmm.lml(), mvn(beta, v0, v1, y, X, K0))

    delta = random.uniform(size=1).item()
    lmm.delta = delta
    v0 = lmm.v0
    v1 = lmm.v1
    assert_allclose(lmm.lml(), mvn(beta, v0, v1, y, X, K0))

    scale = random.random(size=1).item()
    lmm.scale = scale
    v0 = lmm.v0
    v1 = lmm.v1
    assert_allclose(lmm.lml(), mvn(beta, v0, v1, y, X, K0))

    def fun0(x):
        beta = x[:c]
        v0 = exp(x[c])
        v1 = exp(x[c + 1])
        return -mvn(beta, v0, v1, y, X, K0)

    res = minimize(fun0, [0] * c + [0, 0])
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -res.fun, rtol=1e-3, atol=1e-6)
    assert_allclose(lmm.beta, res.x[:c], rtol=1e-3, atol=1e-6)
    assert_allclose(lmm.v0, exp(res.x[c]), rtol=1e-3, atol=1e-6)
    assert_allclose(lmm.v1, exp(res.x[c + 1]), rtol=1e-3, atol=1e-6)

    lmm = LMM(y, X, QS, restricted=restricted)
    beta = random.normal(size=c)
    lmm.beta = beta
    lmm.delta = random.random(size=1).item()
    lmm.scale = random.random(size=1).item()
    lmm.fix("beta")

    def fun1(x):
        v0 = exp(x[0])
        v1 = exp(x[1])
        return -mvn(beta, v0, v1, y, X, K0)

    res = minimize(fun1, [0, 0])
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -res.fun, rtol=1e-3, atol=1e-6)
    assert_allclose(lmm.v0, exp(res.x[0]), rtol=1e-3, atol=1e-6)
    assert_allclose(lmm.v1, exp(res.x[1]), rtol=1e-3, atol=1e-6)

    lmm = LMM(y, X, QS, restricted=restricted)
    lmm.beta = random.normal(size=c)
    delta = random.random(size=1).item()
    lmm.delta = delta
    lmm.scale = random.random(size=1).item()
    lmm.fix("delta")

    def fun2(x):
        beta = x[:c]
        scale = exp(x[c])
        v0 = scale * (1 - delta)
        v1 = scale * delta
        return -mvn(beta, v0, v1, y, X, K0)

    res = minimize(fun2, [0] * c + [0])
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -res.fun, rtol=1e-5, atol=1e-6)
    assert_allclose(lmm.beta, res.x[:c], rtol=1e-5, atol=1e-6)
    assert_allclose(lmm.scale, exp(res.x[c]), rtol=1e-5, atol=1e-6)

    lmm = LMM(y, X, QS, restricted=restricted)
    lmm.beta = random.normal(size=c)
    lmm.delta = random.random(size=1).item()
    scale = random.random(size=1).item()
    lmm.scale = scale
    lmm.fix("scale")

    def fun3(x):
        beta = x[:c]
        delta = 1 / (1 + exp(-x[c]))
        v0 = scale * (1 - delta)
        v1 = scale * delta
        return -mvn(beta, v0, v1, y, X, K0)

    res = minimize(fun3, [0] * c + [0])
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -res.fun, rtol=1e-5, atol=1e-6)
    assert_allclose(lmm.beta, res.x[:c], rtol=1e-3, atol=1e-6)
    assert_allclose(lmm.delta, 1 / (1 + exp(-res.x[c])), rtol=1e-3, atol=1e-6)


def _get_mvn(y, X, G):
    def mvn(beta, v0, v1, y, X, K0):
        m = X @ beta
        K = v0 * K0 + v1 * eye(K0.shape[0])
        return st.multivariate_normal(m, K).logpdf(y)

    return mvn


def _get_mvn_restricted(y, X, G):
    SVD = economic_svd(X)
    tX = ddot(SVD[0], SVD[1])

    def mvn_restricted(beta, v0, v1, y, X, K0):
        n = K0.shape[0]
        m = X @ beta
        K = v0 * K0 + v1 * eye(K0.shape[0])
        lml = st.multivariate_normal(m, K).logpdf(y)
        lml += slogdet(tX.T @ tX)[1] / 2 - slogdet(tX.T @ solve(K, tX))[1] / 2
        lml += n * log(2 * pi) / 2
        lml -= (n - tX.shape[1]) * log(2 * pi) / 2
        return lml

    return mvn_restricted


def _full_rank(random):
    n = 30
    G = random.normal(size=(n, n + 1))
    X = random.normal(size=(n, 2))
    y = (
        X @ random.normal(size=2)
        + G @ random.normal(size=G.shape[1])
        + random.normal(size=n)
    )
    y -= y.mean(0)
    y /= y.std(0)

    return (y, X, G)


def _low_rank(random):
    n = 30
    G = random.normal(size=(n, 5))
    X = random.normal(size=(n, 2))
    X = concatenate((X, X), axis=1)
    y = (
        X @ random.normal(size=4)
        + G @ random.normal(size=G.shape[1])
        + random.normal(size=n)
    )
    y -= y.mean(0)
    y /= y.std(0)
    return (y, X, G)



from numpy import asarray
from numpy.linalg import solve


class LMMPredict:
    """
    Predict new samples.

    Warning
    -------
    This has not been thoroughly tested and the interface is likely to change.
    """

    def __init__(self, y, beta, v0, v1, mean, covariance):
        self._y = y
        self._beta = beta
        self._v0 = v0
        self._v1 = v1
        self._mean = mean
        self._covariance = covariance

    def predictive_mean(self, Xstar, ks, kss):
        mstar = self.mean_star(Xstar)
        ks = self.covariance_star(ks)
        m = self._mean
        K = self._covariance
        return mstar + ks @ solve(K, self._y - m)

    def predictive_covariance(self, Xstar, ks, kss):
        kss = self.variance_star(kss)
        ks = self.covariance_star(ks)
        K = self._covariance
        ktk = solve(K, ks.T)
        b = []
        for i in range(len(kss)):
            b += [ks[i, :] @ ktk[:, i]]
        b = asarray(b)
        return kss - b

    def mean_star(self, Xstar):
        return Xstar @ self._beta

    def variance_star(self, kss):
        return kss * self._v0 + self._v1

    def covariance_star(self, ks):
        return ks * self._v0



"""
Linear mixed models.
"""
from ._kron2sum import Kron2Sum
from ._kron2sum_scan import KronFastScanner
from ._lmm import LMM
from ._lmm_predict import LMMPredict
from ._lmm_scan import FastScanner

__all__ = ["LMM", "FastScanner", "Kron2Sum", "KronFastScanner", "LMMPredict"]



from . import _numbers as numbers
from ._array import unvec, vec
from ._assert import assert_interface
from ._numbers import safe_log
from ._svd import SVD
from .cache import cached_property
from .check import check_covariates, check_economic_qs, check_outcome
from .eigen import economic_qs_zeros
from .format import format_function
from .random import multivariate_normal
from .solve import hinv, hsolve, lu_factor, lu_solve, nice_inv, rsolve

log2pi = 1.837877066409345339081937709124758839607238769531250

__all__ = [
    "SVD",
    "assert_interface",
    "cached_property",
    "check_covariates",
    "check_economic_qs",
    "check_outcome",
    "economic_qs_zeros",
    "format_function",
    "hinv",
    "hsolve",
    "lu_factor",
    "lu_solve",
    "multivariate_normal",
    "nice_inv",
    "numbers",
    "rsolve",
    "safe_log",
    "unvec",
    "vec",
]



from numpy import exp, log
from optimix import Function, Scalar

from .._util import format_function


class LinearCov(Function):
    """
    Linear covariance function, K = s⋅XXᵀ.

    The mathematical representation is s⋅XXᵀ, for a n×r matrix X provided by the user
    and a scaling parameter s.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import LinearCov
        >>> from numpy import dot
        >>> from numpy.random import default_rng
        >>>
        >>> X = default_rng(0).randn(2, 3)
        >>> cov = LinearCov(X)
        >>> cov.scale = 1.3
        >>> cov.name = "K"
        >>> print(cov)
        LinearCov(): K
          scale: 1.3
    """

    def __init__(self, X):
        """
        Constructor.

        Parameters
        ----------
        X : array_like
            Matrix X from K = s⋅XXᵀ.
        """
        self._logscale = Scalar(0.0)
        self._X = X
        Function.__init__(self, "LinearCov", logscale=self._logscale)
        self._logscale.bounds = (-20.0, +10)

    @property
    def X(self):
        """
        Matrix X from K = s⋅XXᵀ.
        """
        return self._X

    def fix(self):
        """
        Prevent s update during optimization.
        """
        self._fix("logscale")

    def unfix(self):
        """
        Enable s update during optimization.
        """
        self._unfix("logscale")

    @property
    def scale(self):
        """
        Scale parameter.
        """
        return exp(self._logscale.value)

    @scale.setter
    def scale(self, scale):
        from numpy_sugar import epsilon

        scale = max(scale, epsilon.tiny)
        self._logscale.value = log(scale)

    def value(self):
        """
        Covariance matrix.

        Returns
        -------
        K : ndarray
            s⋅XXᵀ.
        """
        X = self.X
        return self.scale * (X @ X.T)

    def gradient(self):
        """
        Derivative of the covariance matrix over log(s).

        Returns
        -------
        logscale : ndarray
            s⋅XXᵀ.
        """
        return dict(logscale=self.value())

    def __str__(self):
        return format_function(self, {}, [("scale", self.scale)])



import numpy as np
import pytest
from numpy.testing import assert_allclose

from glimix_core.cov import GivenCov, LinearCov, SumCov


def test_sumcov():
    random = np.random.default_rng(0)
    X = random.random(size=(3, 2))
    cov_left = LinearCov(X)

    K = random.random(size=(3, 3))
    K = K @ K.T
    cov_right = GivenCov(K)

    cov = SumCov([cov_left, cov_right])
    assert_allclose(cov.value(), cov_left.value() + cov_right.value())
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    cov_left.scale = 0.1
    assert_allclose(cov._check_grad(), 0, atol=1e-5)

    with pytest.raises(ValueError):
        K = random.random(size=(3, 3))
        GivenCov(K)



from numpy import maximum, zeros


class Site(object):
    """EP parameters."""

    def __init__(self, n):
        self.tau = zeros(n)
        self.eta = zeros(n)

    def reset(self):
        self.tau[:] = 0
        self.eta[:] = 0

    def update(self, mean, variance, cav):
        self.tau[:] = maximum(1.0 / variance - cav["tau"], 0)
        self.eta[:] = mean / variance - cav["eta"]



from numpy import ascontiguousarray, atleast_1d, atleast_2d, sqrt, std

from ..cov import EyeCov, LinearCov, SumCov
from ..lik import BernoulliProdLik, BinomialProdLik, PoissonProdLik
from ..link import LogitLink, LogLink
from ..mean import LinearMean, OffsetMean, SumMean
from ._ggp import GGPSampler


def bernoulli_sample(
    offset,
    G,
    heritability=0.5,
    causal_variants=None,
    causal_variance=0,
    random_state=None,
):
    r"""Bernoulli likelihood sampling.

    Sample according to

    .. math::

        \mathbf y \sim \prod_{i=1}^n
        \text{Bernoulli}(\mu_i = \text{logit}(z_i))
        \mathcal N(~ o \mathbf 1 + \mathbf a^\intercal \boldsymbol\alpha;
        ~ (h^2 - v_c)\mathrm G^\intercal\mathrm G +
        (1-h^2-v_c)\mathrm I ~)

    using the canonical Logit link function to define the conditional Bernoulli
    mean :math:`\mu_i`.

    The causal :math:`\mathbf a` covariates and the corresponding effect-sizes
    are randomly draw according to the following idea. The ``causal_variants``,
    if given, are first mean-zero and std-one normalized and then having
    its elements divided by the squared-root the the number of variances::

        causal_variants = _stdnorm(causal_variants, axis=0)
        causal_variants /= sqrt(causal_variants.shape[1])

    The causal effect-sizes :math:`\boldsymbol\alpha` are draw from
    :math:`\{-1, +1\}` and subsequently normalized for mean-zero and std-one""

    Parameters
    ----------
    random_state : random_state
        Set the initial random state.

    Example
    -------

    .. doctest::

        >>> from glimix_core.random import bernoulli_sample
        >>> from numpy.random import default_rng
        >>> offset = 5
        >>> G = [[1, -1], [2, 1]]
        >>> bernoulli_sample(offset, G, random_state=default_rng(0))
        array([1., 1.])
    """
    link = LogitLink()
    mean, cov = _mean_cov(
        offset, G, heritability, causal_variants, causal_variance, random_state
    )
    lik = BernoulliProdLik(link)
    sampler = GGPSampler(lik, mean, cov)

    return sampler.sample(random_state)


def binomial_sample(
    ntrials,
    offset,
    G,
    heritability=0.5,
    causal_variants=None,
    causal_variance=0,
    random_state=None,
):
    """Binomial likelihood sampling.

    Parameters
    ----------
    random_state : random_state
        Set the initial random state.

    Example
    -------

    .. doctest::

        >>> from glimix_core.random import binomial_sample
        >>> from numpy.random import default_rng
        >>> ntrials = [5, 15]
        >>> offset = 0.5
        >>> G = [[1, -1], [2, 1]]
        >>> binomial_sample(ntrials, offset, G, random_state=default_rng(0))
        array([ 2., 14.])
    """
    link = LogitLink()
    mean, cov = _mean_cov(
        offset, G, heritability, causal_variants, causal_variance, random_state
    )
    lik = BinomialProdLik(ntrials, link)
    sampler = GGPSampler(lik, mean, cov)

    return sampler.sample(random_state)


def poisson_sample(
    offset,
    G,
    heritability=0.5,
    causal_variants=None,
    causal_variance=0,
    random_state=None,
):
    """Poisson likelihood sampling.

    Parameters
    ----------
    random_state : random_state
        Set the initial random state.

    Example
    -------

    .. doctest::

        >>> from glimix_core.random import poisson_sample
        >>> from numpy.random import default_rng
        >>> offset = -0.5
        >>> G = [[0.5, -1], [2, 1]]
        >>> poisson_sample(offset, G, random_state=default_rng(0))
        array([0., 6.])
    """
    mean, cov = _mean_cov(
        offset, G, heritability, causal_variants, causal_variance, random_state
    )
    link = LogLink()
    lik = PoissonProdLik(link)
    sampler = GGPSampler(lik, mean, cov)

    return sampler.sample(random_state)


def _causal_mean(causal_variants, causal_variance, random):
    causal_variants = atleast_2d(atleast_1d(causal_variants).T).T
    causal_variants = _stdnorm(causal_variants, axis=0)
    causal_variants /= sqrt(causal_variants.shape[1])
    p = causal_variants.shape[1]
    directions = random.normal(size=p)
    directions[directions < 0.5] = -1
    directions[directions >= 0.5] = +1
    s = std(directions)
    if s > 0:
        directions /= s
    directions *= sqrt(causal_variance)
    directions -= directions.mean()
    mean = LinearMean(causal_variants)
    mean.effsizes = directions
    return mean


def _mean_cov(offset, G, heritability, causal_variants, causal_variance, random_state):
    G = ascontiguousarray(G, dtype=float)
    nsamples = G.shape[0]
    G = _stdnorm(G, axis=0)

    G /= sqrt(G.shape[1])

    mean1 = OffsetMean(nsamples)
    mean1.offset = offset

    cov1 = LinearCov(G)
    cov2 = EyeCov(nsamples)
    cov = SumCov([cov1, cov2])

    cov1.scale = heritability - causal_variance
    cov2.scale = 1 - heritability - causal_variance

    means = []
    means.append(mean1)
    if causal_variants is not None:
        means.append(_causal_mean(causal_variants, causal_variance, random_state))

    mean = SumMean(means)

    return mean, cov


def _stdnorm(X, axis=None, out=None):
    X = ascontiguousarray(X, dtype=float)
    if out is None:
        out = X.copy()

    m = out.mean(axis)
    s = out.std(axis)
    ok = s > 0

    out -= m

    out[..., ok] /= s[ok]

    return out



def assert_interface(cls, callables, properties):
    from .cache import cached_property

    attrs = dir(cls)
    private = set(a for a in attrs if a.startswith("_"))
    public = set(attrs) - private

    public_methods = [a for a in public if callable(getattr(cls, a))]
    public_props = [
        a for a in public if isinstance(getattr(cls, a), (property, cached_property))
    ]

    _assert_callables(set(public_methods), callables)
    _assert_properties(set(public_props), properties)


def _assert_callables(public, required):
    missing = public - set(required)
    if missing:
        msg = "The following public methods exist but have not been asserted: "
        msg += "{}".format(", ".join(list(missing)))
        raise AssertionError(msg)

    missing = set(required) - public
    if missing:
        msg = "The following methods have not been found: "
        msg += "{}".format(", ".join(list(missing)))
        raise AssertionError(msg)


def _assert_properties(public, required):
    missing = public - set(required)
    if missing:
        msg = "The following public properties exist but have not been asserted: "
        msg += "{}".format(", ".join(list(missing)))
        raise AssertionError(msg)

    missing = set(required) - public
    if missing:
        msg = "The following properties have not been found: "
        msg += "{}".format(", ".join(list(missing)))
        raise AssertionError(msg)



from numpy import matmul, sqrt
from numpy.random import default_rng
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.ggp import ExpFamGP
from glimix_core.lik import BernoulliProdLik
from glimix_core.link import LogitLink
from glimix_core.mean import OffsetMean


def _get_data():
    random = default_rng(1)
    N = 10
    X = random.normal(size=(N, N + 1))
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean(N)
    mean.offset = offset

    cov_left = LinearCov(X)
    cov_left.scale = 1.5

    cov_right = EyeCov(N)
    cov_right.scale = 1.5

    cov = SumCov([cov_left, cov_right])

    lik = BernoulliProdLik(LogitLink())

    # y = GGPSampler(lik, mean, cov).sample(random)
    y = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0]
    return dict(
        mean=mean, cov=cov, lik=lik, y=y, cov_left=cov_left, cov_right=cov_right
    )


def test_ggp_expfam():
    data = _get_data()
    ep = ExpFamGP(data["y"], "bernoulli", data["mean"], data["cov"])
    assert_allclose(ep.value(), -4.905982177317817)
    assert_allclose(ep._check_grad(), 0, atol=1e-4)
    data["cov_left"].fix()
    ep.fit(verbose=False)
    assert_allclose(data["cov_right"].scale, 1.133659133003428e-08, atol=1e-5)
    assert_allclose(data["mean"].offset, 2.7591672948875834, rtol=1e-6)


def test_ggp_expfam_tobi():
    random = default_rng(2)

    n = 30

    ntrials = random.integers(30, size=n)
    K = random.normal(size=(n, n))
    K = matmul(K, K.T)

    # lik = BinomialProdLik(ntrials=ntrials, link=LogitLink())

    mean = OffsetMean(n)

    cov2 = EyeCov(n)

    # y = GGPSampler(lik, mean, cov2).sample(random)
    y = [
        19.0,
        4.0,
        1.0,
        4.0,
        4.0,
        21.0,
        8.0,
        1.0,
        5.0,
        6.0,
        17.0,
        5.0,
        10.0,
        1.0,
        7.0,
        0.0,
        7.0,
        3.0,
        4.0,
        18.0,
        3.0,
        0.0,
        4.0,
        2.0,
        18.0,
        12.0,
        14.0,
        0.0,
        12.0,
        3.0,
    ]

    ggp = ExpFamGP(y, ("binomial", ntrials), mean, cov2)
    assert_allclose(ggp.lml(), -75.92472423326483)

    ggp.fit(verbose=False)
    assert_allclose(ggp.lml(), -75.08008540883655)



import os
import sys

import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath(".."))


def _get_version():
    import glimix_core

    return glimix_core.__version__


def _get_name():
    import glimix_core

    return glimix_core.__name__


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

autodoc_default_flags = ["members"]
autosummary_generate = True
napoleon_numpy_docstring = True
templates_path = ["_templates"]

source_suffix = ".rst"

master_doc = "index"

project = _get_name()
copyright = "2018, Danilo Horta"
author = "Danilo Horta"

version = _get_version()
release = version

language = None

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "conf.py"]

pygments_style = "default"
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_sidebars = {"**": ["relations.html", "searchbox.html"]}

htmlhelp_basename = "{}doc".format(project)

man_pages = [(master_doc, _get_name(), "{} documentation".format(project), [author], 1)]

intersphinx_mapping = {
    "python": ("http://docs.python.org/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference/", None),
}



def _nsamples():
    return 10


def offset_mean():
    from glimix_core.mean import OffsetMean

    mean = OffsetMean(_nsamples())
    mean.offset = 0.5

    return mean


def linear_eye_cov():
    from numpy import sqrt
    from numpy.random import default_rng

    from glimix_core.cov import EyeCov, LinearCov, SumCov

    random = default_rng(458)
    X = random.normal(size=(_nsamples(), _nsamples() + 1))
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])

    cov_left = LinearCov(X)
    cov_left.scale = 1.0

    cov_right = EyeCov(_nsamples())
    cov_right.scale = 1.0

    return SumCov([cov_left, cov_right])



from __future__ import division

from numpy import array
from numpy.testing import assert_allclose

from glimix_core.cov import LRFreeFormCov


def test_lrfreeformcov():
    cov = LRFreeFormCov(3, 2)
    L = array([[1, 0], [2, 1], [-1, 3]], float)
    cov.L = L
    assert_allclose(cov.value(), L @ L.T)
    assert_allclose(cov._check_grad(), 0, atol=1e-5)



import scipy.stats as st
from brent_search import minimize
from numpy import concatenate, empty, kron
from numpy.random import default_rng
from numpy.testing import assert_allclose
from numpy_sugar.linalg import economic_qs

from glimix_core._util import assert_interface, vec
from glimix_core.lmm import FastScanner, Kron2Sum, KronFastScanner


def test_lmm_kron_scan():
    random = default_rng(0)
    n = 20
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 6))
    lmm = Kron2Sum(Y, A, F, G, restricted=True)
    lmm.fit(verbose=False)
    scan = lmm.get_fast_scanner()

    m = lmm.mean()
    K = lmm.covariance()

    def func0(scale):
        mv = st.multivariate_normal(m, scale * K)
        return -mv.logpdf(vec(Y))

    s = minimize(func0, 1e-3, 5.0, 1e-5)[0]

    assert_allclose(scan.null_lml(), st.multivariate_normal(m, s * K).logpdf(vec(Y)))
    assert_allclose(kron(A, F) @ scan.null_beta, m)

    A1 = random.normal(size=(3, 2))
    F1 = random.normal(size=(n, 4))

    r = scan.scan(A1, F1)
    assert_allclose(r["scale"], 0.7432575497, rtol=1e-3)

    m = kron(A, F) @ vec(r["effsizes0"]) + kron(A1, F1) @ vec(r["effsizes1"])

    def func1(scale):
        mv = st.multivariate_normal(m, scale * K)
        return -mv.logpdf(vec(Y))

    s = minimize(func1, 1e-3, 5.0, 1e-5)[0]

    assert_allclose(r["lml"], st.multivariate_normal(m, s * K).logpdf(vec(Y)))

    r = scan.scan(empty((3, 0)), F1)
    assert_allclose(r["lml"], -67.2013784411, rtol=1e-4)
    assert_allclose(r["scale"], 0.8999995537936586, rtol=1e-3)
    assert_allclose(
        r["effsizes0"],
        [
            [-0.1463056513, -0.8234240129, -0.1653985457],
            [0.164938537, 1.0066226167, 0.1740701106],
        ],
        rtol=1e-2,
        atol=1e-2,
    )
    assert_allclose(r["effsizes1"], [])


def test_lmm_kron_scan_with_lmm():
    random = default_rng(0)
    n = 15
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 6))

    klmm = Kron2Sum(Y, A, F, G, restricted=True)
    klmm.fit(verbose=False)
    kscan = klmm.get_fast_scanner()

    K = klmm.covariance()

    X = kron(A, F)
    QS = economic_qs(K)
    scan = FastScanner(vec(Y), X, QS, 0.0)

    assert_allclose(klmm.covariance(), K)
    assert_allclose(kscan.null_scale, scan.null_scale)
    assert_allclose(kscan.null_beta, scan.null_beta)
    assert_allclose(kscan.null_lml(), scan.null_lml())
    assert_allclose(kscan.null_beta_covariance, scan.null_beta_covariance)

    A1 = random.normal(size=(3, 2))
    F1 = random.normal(size=(n, 2))
    M = kron(A1, F1)

    kr = kscan.scan(A1, F1)
    r = scan.scan(M)
    assert_allclose(kr["lml"], r["lml"])
    assert_allclose(kr["scale"], r["scale"])
    assert_allclose(vec(kr["effsizes0"]), r["effsizes0"])
    assert_allclose(vec(kr["effsizes1"]), r["effsizes1"])
    assert_allclose(vec(kr["effsizes0_se"]), r["effsizes0_se"])
    assert_allclose(vec(kr["effsizes1_se"]), r["effsizes1_se"])


def test_lmm_kron_scan_unrestricted():
    random = default_rng(0)
    n = 15
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 6))
    lmm = Kron2Sum(Y, A, F, G, restricted=False)
    lmm.fit(verbose=False)
    scan = lmm.get_fast_scanner()

    assert_allclose(scan.null_scale, 1.0, rtol=1e-3)
    assert_allclose(lmm.beta_covariance, scan.null_beta_covariance, rtol=1e-3)


def test_lmm_kron_scan_redundant():
    random = default_rng(0)
    n = 30
    Y = random.normal(size=(n, 3))
    A = random.normal(size=(3, 3))
    A = A @ A.T
    F = random.normal(size=(n, 2))
    G = random.normal(size=(n, 6))
    G = concatenate([G, G], axis=1)
    lmm = Kron2Sum(Y, A, F, G, restricted=True)
    lmm.fit(verbose=False)
    scan = lmm.get_fast_scanner()

    m = lmm.mean()
    K = lmm.covariance()

    def func0(scale):
        mv = st.multivariate_normal(m, scale * K)
        return -mv.logpdf(vec(Y))

    s = minimize(func0, 1e-3, 5.0, 1e-5)[0]

    assert_allclose(scan.null_lml(), st.multivariate_normal(m, s * K).logpdf(vec(Y)))
    assert_allclose(kron(A, F) @ scan.null_beta, m)

    A1 = random.normal(size=(3, 2))
    F1 = random.normal(size=(n, 4))
    F1 = concatenate([F1, F1], axis=1)

    r = scan.scan(A1, F1)
    assert_allclose(r["scale"], 0.8558704225, rtol=1e-3)

    m = kron(A, F) @ vec(r["effsizes0"]) + kron(A1, F1) @ vec(r["effsizes1"])

    def func1(scale):
        mv = st.multivariate_normal(m, scale * K)
        return -mv.logpdf(vec(Y))

    s = minimize(func1, 1e-3, 5.0, 1e-5)[0]

    assert_allclose(r["lml"], st.multivariate_normal(m, s * K).logpdf(vec(Y)))


def test_lmm_kron_scan_public_attrs():
    assert_interface(
        KronFastScanner,
        ["null_lml", "scan"],
        ["null_beta", "null_beta_covariance", "null_beta_se", "null_scale"],
    )



class SVD:
    """
    Represents a SVD decomposition with some useful methods.

    Let 𝙰 be a given 𝑚×𝑛 real matrix. The SVD decomposition of 𝙰 is defined by ::

        𝙰 = 𝚄𝚂𝚅ᵀ

    for which 𝚂 is a diagonal matrix.

    Parameters
    ----------
    A
        Matrix 𝙰. Optional if the parameter `USVt` is given.
    USVt
        SVD decomposition of 𝙰. Optional if the parameter `A` is given.
    """

    def __init__(self, A=None, USVt=None):
        from numpy_sugar.linalg import ddot, economic_svd

        if A is None and USVt is None:
            raise ValueError("Both `A` and `USVt` cannot be `None`.")

        if A is None:
            self._US = ddot(USVt[0], USVt[1])
            self._Vt = USVt[2]
            self._A = self._US @ self._Vt
        else:
            USVt = economic_svd(A)
            self._US = ddot(USVt[0], USVt[1])
            self._Vt = USVt[2]
            self._A = A

        self._rank = len(USVt[1])

    @property
    def A(self):
        """
        Get 𝙰.
        """
        return self._A

    @property
    def US(self):
        """
        Get 𝚄𝚂.
        """
        return self._US

    @property
    def Vt(self):
        """
        Get 𝚅ᵀ.
        """
        return self._Vt

    @property
    def rank(self) -> int:
        """
        Get the diagonal size of matrix 𝚂.
        """
        return self._rank



import warnings

from numpy import abs as npy_abs
from numpy import (
    absolute,
    arctan2,
    array,
    atleast_1d,
    cos,
    divide,
    errstate,
    full_like,
    isfinite,
    logical_and,
    logical_not,
    maximum,
    minimum,
    nan_to_num,
    sign,
    sin,
    sqrt,
    zeros,
    equal,
    hypot,
)
from numpy.linalg import LinAlgError, pinv


def lu_solve(A, b, *args, **kwargs):
    from scipy.linalg import lu_solve

    if A[0].shape[1] == 0 and b.shape[0] == 0:
        return zeros((A[0].shape[0], b.shape[-1]))

    return lu_solve(A, b, *args, **kwargs)


def lu_factor(A, *args, **kwargs):
    from scipy.linalg import lu_factor

    if sum(A.shape) == 0:
        return (zeros((0, 0)), zeros((0,), dtype="int32"))

    return lu_factor(A, *args, **kwargs)


def nice_inv(A):
    """
    Nice inverse.
    """
    from numpy_sugar.linalg import sum2diag

    return pinv(sum2diag(A, 1e-12))


def rsolve(A, y):
    """
    Robust solve Ax=y.
    """
    from numpy_sugar.linalg import rsolve as _rsolve

    try:
        beta = _rsolve(A, y)
    except LinAlgError:
        msg = "Could not converge to solve Ax=y."
        msg += " Setting x to zero."
        warnings.warn(msg, RuntimeWarning)
        beta = zeros(A.shape[0])

    return beta


def heigvals(a, b, d):
    a = float(a)
    b = atleast_1d(b)
    d = atleast_1d(d)
    T = a + d
    D = a * d - b * b

    t0 = T / 2
    t1 = sqrt(maximum(T * T / 4 - D, 0))
    eig0 = t0 + t1
    eig1 = t0 - t1

    eig0 = absolute(eig0)
    eig1 = absolute(eig1)

    with errstate(invalid="ignore", divide="ignore"):
        return nan_to_num(maximum(eig0, eig1) / minimum(eig0, eig1))


def _hinv_svd(a, b, d):
    rcond = 1e-15
    U, S, VT = hsvd(a, b, d)

    maxi = maximum(absolute(S[0]), absolute(S[1]))
    cutoff = rcond * maxi

    large = S[0] > cutoff
    S[0] = divide(1, S[0], where=large, out=S[0])
    S[0][~large] = 0

    large = S[1] > cutoff
    S[1] = divide(1, S[1], where=large, out=S[1])
    S[1][~large] = 0

    SiVT = [[VT[0][0] * S[0], VT[0][1] * S[0]], [VT[1][0] * S[1], VT[1][1] * S[1]]]
    Ai = [
        [
            U[0][0] * SiVT[0][0] + U[0][1] * SiVT[1][0],
            U[0][0] * SiVT[0][1] + U[0][1] * SiVT[1][1],
        ],
        [
            U[1][0] * SiVT[0][0] + U[1][1] * SiVT[1][0],
            U[1][0] * SiVT[0][1] + U[1][1] * SiVT[1][1],
        ],
    ]
    ai = Ai[0][0]
    bi = Ai[0][1]
    di = Ai[1][1]

    return ai, bi, di


def hinv(a, b, d):
    rcond = 1e-7

    b = atleast_1d(b)
    d = atleast_1d(d)
    a = float(a)
    cond = heigvals(a, b, d)
    norm = a * d - b * b
    with errstate(invalid="ignore", divide="ignore"):
        ai = d / norm
        bi = -b / norm
        di = a / norm

    nok = cond > 1 / rcond
    ai[nok], bi[nok], di[nok] = _hinv_svd(a, b[nok], d[nok])
    return ai, bi, di


def _hinv(A00, A01, A11):
    from numpy_sugar import is_all_finite

    rcond = 1e-15
    b = atleast_1d(A01)
    d = atleast_1d(A11)
    a = full_like(d, A00)
    m = maximum(maximum(npy_abs(b), npy_abs(d)), abs(a))

    a /= m
    b = b / m
    c = b
    d = d / m

    bc = b * c
    ad = a * d
    with errstate(invalid="ignore", divide="ignore"):
        ai = a / (a * a - nan_to_num((bc * a) / d))
        bi = b / (b * b - nan_to_num(ad))
        di = d / (d * d - nan_to_num((bc * d) / a))

    ai /= m
    bi /= m
    di /= m

    ok = is_all_finite(ai) and is_all_finite(bi) and is_all_finite(di)
    if not ok:
        ok = logical_and.reduce([isfinite(ai), isfinite(bi), isfinite(di)])
        nok = logical_not(ok)
        U, S, VT = hsvd(a[nok], b[nok], d[nok])

        maxi = maximum(npy_abs(S[0]), npy_abs(S[1]))
        cutoff = rcond * maxi

        large = S[0] > cutoff
        S[0] = divide(1, S[0], where=large, out=S[0])
        S[0][~large] = 0

        large = S[1] > cutoff
        S[1] = divide(1, S[1], where=large, out=S[1])
        S[1][~large] = 0

        SiVT = [[VT[0][0] * S[0], VT[0][1] * S[0]], [VT[1][0] * S[1], VT[1][1] * S[1]]]
        Ai = [
            [
                U[0][0] * SiVT[0][0] + U[0][1] * SiVT[1][0],
                U[0][0] * SiVT[0][1] + U[0][1] * SiVT[1][1],
            ],
            [
                U[1][0] * SiVT[0][0] + U[1][1] * SiVT[1][0],
                U[1][0] * SiVT[0][1] + U[1][1] * SiVT[1][1],
            ],
        ]
        ai[nok] = Ai[0][0] / m
        bi[nok] = Ai[0][1] / m
        di[nok] = Ai[1][1] / m

    return ai, bi, di


def hsvd_old(a, b, d):
    a = atleast_1d(a)
    b = atleast_1d(b)
    d = atleast_1d(d)

    aa = a * a
    bb = b * b
    dd = d * d
    ab = a * b
    bd = b * d

    e = aa - dd
    s1 = aa + 2 * bb + dd
    s2 = sqrt(e**2 + 4 * (ab + bd) ** 2)

    t = 2 * ab + 2 * bd
    theta = arctan2(t, e) / 2
    psi = arctan2(t, aa - dd) / 2

    Ct = cos(theta)
    St = sin(theta)
    Cp = cos(psi)
    Sp = sin(psi)

    s11 = (a * Ct + b * St) * Cp + (b * Ct + d * St) * Sp
    s22 = (a * St - b * Ct) * Sp + (-b * St + d * Ct) * Cp

    U = [[Ct, -St], [St, Ct]]
    S = [sqrt((s1 + s2) / 2), sqrt(maximum((s1 - s2) / 2, 0.0))]

    VT = [[sign(s11) * Cp, sign(s11) * Sp], [-sign(s22) * Sp, sign(s22) * Cp]]

    # U S V.T
    return U, S, VT


def hsvd(a, b, d):
    a = atleast_1d(a)
    b = atleast_1d(b)
    d = atleast_1d(d)
    y1, x1 = 2 * b, (a - d)
    y2, x2 = 0, (a + d)

    h1 = hypot(y1, x1)
    h2 = hypot(y2, x2)

    with errstate(invalid="ignore", divide="ignore"):
        t1 = x1 / h1
        t2 = x2 / h2

    t1[equal(y1, 0)] = 1.0
    t2[equal(y2, 0)] = 1.0

    cc = sqrt((1 + t1) * (1 + t2))
    ss = sqrt((1 - t1) * (1 - t2))
    cs = sqrt((1 + t1) * (1 - t2))
    sc = sqrt((1 - t1) * (1 + t2))

    c1, s1 = (
        (cc - ss) / 2,
        (sc + cs) / 2,
    )
    U = [[c1, -s1], [s1, c1]]

    S = [(h1 + h2) / 2, (h1 - h2) / 2]

    VT = [[c1 * a + s1 * b, c1 * b + s1 * d], [-s1 * a + c1 * b, -s1 * b + c1 * d]]

    with errstate(invalid="ignore", divide="ignore"):
        VT[0][0] /= S[0]
        VT[0][1] /= S[0]
        VT[1][0] /= S[1]
        VT[1][1] /= S[1]

    eq = equal(h1 + h2, 0)
    VT[0][0][eq] = 0
    VT[0][1][eq] = 0

    eq = equal(h1, h2)
    VT[1][0][eq] = 0
    VT[1][1][eq] = 0

    return U, S, VT


def hsolve(A00, A01, A11, y0, y1):
    """
    Solver for the linear equations of two variables and equations only.

    It uses Householder reductions to solve A𝐱 = 𝐲 in a robust manner.

    Parameters
    ----------
    A : array_like
        Coefficient matrix.
    y : array_like
        Ordinate values.

    Returns
    -------
    ndarray
        Solution 𝐱.
    """
    from numpy_sugar import epsilon

    n = _norm(A00, A01)
    u0 = A00 - n
    u1 = A01
    nu = _norm(u0, u1)

    with errstate(invalid="ignore", divide="ignore"):
        v0 = nan_to_num(u0 / nu)
        v1 = nan_to_num(u1 / nu)

    B00 = 1 - 2 * v0 * v0
    B01 = 0 - 2 * v0 * v1
    B11 = 1 - 2 * v1 * v1

    D00 = B00 * A00 + B01 * A01
    D01 = B00 * A01 + B01 * A11
    D11 = B01 * A01 + B11 * A11

    b0 = y0 - 2 * y0 * v0 * v0 - 2 * y1 * v0 * v1
    b1 = y1 - 2 * y0 * v1 * v0 - 2 * y1 * v1 * v1

    n = _norm(D00, D01)
    u0 = D00 - n
    u1 = D01
    nu = _norm(u0, u1)

    with errstate(invalid="ignore", divide="ignore"):
        v0 = nan_to_num(u0 / nu)
        v1 = nan_to_num(u1 / nu)

    E00 = 1 - 2 * v0 * v0
    E01 = 0 - 2 * v0 * v1
    E11 = 1 - 2 * v1 * v1

    F00 = E00 * D00 + E01 * D01
    F01 = E01 * D11
    F11 = E11 * D11

    F11 = (npy_abs(F11) > epsilon.small) * F11

    with errstate(divide="ignore", invalid="ignore"):
        Fi00 = nan_to_num(F00 / F00 / F00)
        Fi11 = nan_to_num(F11 / F11 / F11)
        Fi10 = nan_to_num(-(F01 / F00) * Fi11)

    c0 = Fi00 * b0
    c1 = Fi10 * b0 + Fi11 * b1

    x0 = E00 * c0 + E01 * c1
    x1 = E01 * c0 + E11 * c1

    return array([x0, x1])


def _norm(x0, x1):
    m = maximum(npy_abs(x0), npy_abs(x1))
    with errstate(invalid="ignore"):
        a = (x0 / m) * (x0 / m)
        b = (x1 / m) * (x1 / m)
        return nan_to_num(m * sqrt(a + b))



r"""
****************************
Generalised Gaussian Process
****************************

Introduction
^^^^^^^^^^^^

A GP is a statistical distribution :math:`Y_t`, :math:`t\in\mathrm T`, for
which any finite linear combination of samples has a joint Gaussian
distribution `[1]`_ `[2]`_.
An instance of such class of processes is defined via a mean function
:math:`m(\cdot)` and a covariance function :math:`k(\cdot, \cdot)` whose
domains are :math:`\mathrm T` and :math:`\mathrm T\times\mathrm T`,
respectively.

Here we implement an extension of GPs that makes use of exponential-family
likelihoods.
An instance of such process is given by

.. math::

    \mathbf y \sim \int \prod_i \text{ExpFam}(y_i ~|~ \mu_i = g(z_i))
      \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K) \mathrm d\mathbf z.

:class:`.ExpFamGP` performs inference over the mean and covariance
parameters via maximum likelihood and Expectation Propagation `[3]`_
approximation.

.. _[1]: https://en.wikipedia.org/wiki/Gaussian_process
.. _[2]: http://www.gaussianprocess.org/gpml/
.. _[3]: http://www.gaussianprocess.org/gpml/chapters/RW3.pdf

Usage
^^^^^
"""
from ._expfam import ExpFamGP

__all__ = ["ExpFamGP"]



from numpy import kron, ravel
from numpy.random import default_rng
from numpy.testing import assert_allclose

from glimix_core.mean import KronMean


def test_mean_kron():

    random = default_rng(0)
    # number of trais
    p = 2
    # number of covariates
    c = 3
    # sample size
    n = 4

    A = random.normal(size=(p, p))
    X = random.normal(size=(n, c))
    B = random.normal(size=(p, c))

    mean = KronMean(A, X)
    mean.B = B
    assert_allclose(mean.value(), kron(A, X) @ ravel(B, order="F"))
    assert_allclose(mean._check_grad(), 0, atol=1e-5)
